"""
Investigate feature detector "robustness" - the ability to detect the same 
features/speckles in multiple frames for speckle pattern rotation.
"""
#%%-----------------------------------------------------------------------------
# setup
#-------------------------------------------------------------------------------
import lib.detectors as detectors
import lib.misc as misc
import h5py
import numpy as np
#%%-----------------------------------------------------------------------------
# robustness check function
#-------------------------------------------------------------------------------   
def rotation_robustness_match( wins, Dangs, Axs, Ays, detector, radius=100,ref_i=0, dn=1):
    """
    Do robustness check on rotating window data (see get_rotated_windows)
    
    1) Find keypoints in first image window within the radius given to ensure 
        all keypoints will be present when rotated.
    
    2) Find keypoints in a rotated window again within the radius given
       
    3) Calculate the rotated keypoint positions in the reference frame and 
        compare positions of keypoints to match if there is a keypoint within 
        +-dn px
        
    
    Returns results dictionary with keys:

    method  -   method used
    tolerance  -   max distance in px for match to be verified
    ref_i   -   reference frame index used
    radius  -   radius of window used.
    N   -   Number of features found
    Nm  -   Number of feature smatched by descriptor,
    R = Nm/N0 - robustness where N0 = N[ref_i]
    
    """
    res = {}
    res['method'] = detector.method
    res['tolerance'] = dn
    res['ref_i'] = ref_i
    res['radius'] = radius
    
    #reference window
    win0 = wins[ref_i]
    
    #find keypoints in original window
    kps0 = detector.find_features(win0)
    
    #filter keypoints with radius greater than that given
    Cy,Cx  = np.array(win0.shape)//2 #center or rotation
    kps0 = misc.filter_kps( kps0, centre=(Cy,Cx), radius=radius)
    N0 = len(kps0)
    print 'N0 : ',N0
    
    #Loop over all windows and match to first window found
    N = []         #number of keypoints found
    Nm = []   #number matched to +-dn pixel
    for n in range(0,len(wins) ):
        
        #get window and angle delta
        Dang = Dangs[n]
        win1 = wins[n]
        
        #find keypoints in next window
        kps1 = detector.find_features(wins[n])
        
        #filter keypoints with radius greater than that given
        kps1 = misc.filter_kps( kps1, centre=(Cy,Cx), radius=radius)
        N.append(len(kps1))

        print 'Image: ',n,'Dang: ',Dang, 'Number of kps:' ,len(kps1), 

        #rotate the keypoints to reference frame (i.e. -Dang)
        #+ve here due to definition of Dang and image coords
        kps1_ref = misc.rotate_kps( kps1, Dang, centre=(Cy,Cx))
        
        #check each kp0 for match to within +- dn pixels in this rotated set
        matches = misc.proximity_match( kps0, kps1_ref, dn)
        found = len(matches)
        Nm.append( found )
        print 'Matched: ', found
        
    #convert to array type for ease
    N = np.array(N, np.float64)
    Nm = np.array(Nm, np.float64)
    
    #calculate robustness R
    R = Nm/N0
    
    #store results to results dicttionary
    res['N'] = N
    res['Nm'] = Nm
    res['R'] = R
    
    return res
    
#%%-----------------------------------------------------------------------------
# setup samples names etc
#-------------------------------------------------------------------------------
tolerance = 1.5 #tolerance used in proximity matching
ref_i = 0       #reference to this image
radius = 360    #radius of circular window
pad = 32        #extra border padding to allow feature detectors to work upto radius

#surface preparation methods
surfaces = ( ('horizontal milling', (2.5,16,50)),
            ('flat lapping reaming', (0.55,1.6,10)),
            ('vertical milling', (2.5,16,50)),
            ('turning', (2.5,16,50)),
            ('grinding', (0.55,1.6,10))  )
            
surfaces = ( ('horizontal milling', (2.5,)),  )

surfaces = ( ('turning', (50,)),
            ('horizontal milling', (16,)),
            ('flat lapping reaming', (1.6,)),
            ('vertical milling', (16,)),)
            
path = '..\\data\\dataset5 surface samples rotation\\'

out = h5py.File('..\\results\\5.3 detector_rotation_robustness.hdf5', 'a')

#%%-----------------------------------------------------------------------------
# Loop over each sample
#-------------------------------------------------------------------------------
for surface, Rzs in surfaces:
    for Rz in Rzs:
        print '------------------------------ {0} {1} ------------------------------'.format(surface, Rz)

        #-------------------------------------------------------------------------------
        # load from hdf5
        #-------------------------------------------------------------------------------
        filename = '{0} Rz={1} um'.format(surface, Rz)

        print 'Loading data from :', path+filename

        data = h5py.File(path+filename, 'r')
        imgs = data['imgs']
        angs = data['angs'][:]

        #-------------------------------------------------------------------------------
        # get rotated/centered windows
        #-------------------------------------------------------------------------------
        centre=(490,625) 
        Ax, Ay, Dangs, c, wins, wins_ = misc.get_rotated_windows( imgs, angs, centre=centre, radius=radius,pad=pad, ref_i=ref_i, mask=False)
        #wins are centered window but still rotated originals
        #wins_ are centered and rotated to same oreientation for testing comparision

        #-------------------------------------------------------------------------------
        # Fast
        #-------------------------------------------------------------------------------
        print 'Doing FAST'
        detector = detectors.FAST_Detector()
        res_fast = rotation_robustness_match( wins, Dangs, Ax, Ay, detector, radius=radius, ref_i=ref_i, dn=tolerance)

        #%%-----------------------------------------------------------------------------
        # AGAST
        #-------------------------------------------------------------------------------
        print 'Doing AGAST'
        detector = detectors.AGAST_Detector()
        res_agast = rotation_robustness_match( wins, Dangs, Ax, Ay, detector, radius=radius, ref_i=ref_i, dn=tolerance)


        #-------------------------------------------------------------------------------
        # Harris
        #-------------------------------------------------------------------------------
        print 'Doing Harris'

        detector = detectors.Harris_Detector()
        res_harris = rotation_robustness_match( wins, Dangs, Ax, Ay, detector, radius=radius, ref_i=ref_i, dn=tolerance)

        #-------------------------------------------------------------------------------
        # USURF/DOG
        #-------------------------------------------------------------------------------
        print 'Doing USURF/DoG'
        detector = detectors.USURF_Detector()
        res_dog = rotation_robustness_match( wins, Dangs, Ax, Ay, detector, radius=radius, ref_i=ref_i, dn=tolerance)

        #-------------------------------------------------------------------------------
        # SURF/oDOG
        #-------------------------------------------------------------------------------
        print 'Doing SURF/oDoG'
        detector = detectors.SURF_Detector()
        res_odog = rotation_robustness_match( wins, Dangs, Ax, Ay, detector, radius=radius, ref_i=ref_i, dn=tolerance)

        #-------------------------------------------------------------------------------
        # ORB/oFAST
        #-------------------------------------------------------------------------------
        print 'Doing ORB/oFAST'
        detector = detectors.ORB_Detector()
        res_ofast = rotation_robustness_match( wins, Dangs, Ax, Ay, detector, radius=radius, ref_i=ref_i, dn=tolerance)

        #-------------------------------------------------------------------------------
        # BRISK/oAGAST
        #-------------------------------------------------------------------------------
        print 'Doing BRISK/oAGAST'
        detector = detectors.BRISK_Detector()
        res_oagast = rotation_robustness_match( wins, Dangs, Ax, Ay, detector, radius=radius, ref_i=ref_i, dn=tolerance)

        #-------------------------------------------------------------------------------
        # store processed results
        #-------------------------------------------------------------------------------
        surface_full = '{0} Rz={1} um'.format(surface, Rz)

        #store common items to the surface group
        common = {'Ax':Ax, 'Ay':Ay, 'Dangs':Dangs, 'angs':angs, 'c':c, 'tolerance':tolerance }
        surface_group = misc.save_dict_to_hdf5( common , out, key=surface_full)

        #save methods
        misc.save_dict_to_hdf5(res_fast, surface_group, 'FAST')
        misc.save_dict_to_hdf5(res_agast, surface_group, 'AGAST')
        misc.save_dict_to_hdf5(res_harris, surface_group, 'Harris')
        misc.save_dict_to_hdf5(res_dog, surface_group, 'DoG')
        misc.save_dict_to_hdf5(res_odog, surface_group, 'ODoG')
        misc.save_dict_to_hdf5(res_ofast, surface_group, 'OFAST')
        misc.save_dict_to_hdf5(res_oagast, surface_group, 'OAGAST')

out.close()





