"""
Investigate feature detector "robustness" - the ability to detect the same 
features/speckles in multiple frames:


Method/robustness in the presence of translations:
    1) find the shift between frames using normalised cross-correaltion
    2) Find keypoints in first image window defined by rect0. 
    3) Find keypoints in a shifted window in the next image, where the window is
    shifted to account for the speckle shift and ensure that all speckles are 
    still present and only interframe changes/decorrelation of the speckles 
    influences the detection
    4)Compare positions of keypoints to match if there is a keypoint within +-1px

"""
#%%-----------------------------------------------------------------------------
# setup
#-------------------------------------------------------------------------------
import numpy as np
import matplotlib.pyplot as plt
plt.ion()
import time
import cv2 as cv

import lib.detectors as detectors
import lib.misc as misc

import h5py

#%%-----------------------------------------------------------------------------
# do robustness check
#-------------------------------------------------------------------------------  
def robustness_match( wins, detector, dn=1, ref_i=0):
    """
    Do robustness check on shifted windows (see get_shifted_windows).
    
    1) Find keypoints in first image window
    2) Find keypoints in a shifted window
    3) Compare positions of keypoints to match if there is a keypoint within 
        +-dn px
        
    Returns:
    dictionary of setttings/results keys=method, tolerance, ref_i, N, Nm, R
    """
    res = {}
    res['method'] = detector.method
    res['tolerance'] = dn
    res['ref_i'] = ref_i
    
    #find keypoints in original window
    kps0 = detector.find_features(wins[ref_i]) 
    N0 = len(kps0)
    
    #Loop over all windows and match to first window found
    N = []         #number of keypoints found
    Nm = []        #number matched to +-1 pixel

    for n in range(0,len(wins) ):
        
        #find features in next window
        win1 = wins[n]
        kps1 = detector.find_features(win1)
        N.append(len(kps1))
        print 'Image: ',n, 'Number of kps:' , len(kps1), 
        
        #match by proximity
        matches = misc.proximity_match( kps0, kps1, dn)
        found = len(matches)
        Nm.append( found )
        print 'Matched: ', found
        
    #convert to array type for ease
    N = np.array(N, np.float64)
    Nm = np.array(Nm, np.float64)
    
    #calculate robustness R
    R = Nm/N0
    
    #store to results dict
    res['N'] = N
    res['Nm'] = Nm
    res['R'] = R
    
    return res

#%%-----------------------------------------------------------------------------
# setup samples names etc
#-------------------------------------------------------------------------------
tolerance = 1.5 #tolerance used in proximity matching
ref_i = 0       #reference to this image

#surface preparation methods
surfaces = ( ('horizontal milling', (2.5,16,50)),
            ('flat lapping reaming', (0.55,1.6,10)),
            ('vertical milling', (2.5,16,50)),
            ('turning', (2.5,16,50)),
            ('grinding', (0.55,1.6,10))  )

path = '..\\data\\dataset4 surface samples translation\\'

out = h5py.File('results\\5.3 detector_translation_robustness.hdf5', 'a')    
  
#%%-----------------------------------------------------------------------------
# Loop over each sample
#-------------------------------------------------------------------------------
for surface, Rzs in surfaces:
    for Rz in Rzs:
        print '------------------------------ {0} {1} ------------------------------'.format(surface, Rz)

        #%%-----------------------------------------------------------------------------
        # load from hdf5
        #-------------------------------------------------------------------------------
        filename = '{0} Rz={1} um.hdf5'.format(surface, Rz)
        print 'Loading data from :', path+filename

        data = h5py.File(path+filename, 'r')
        imgs = data['imgs']
        
        #%%-----------------------------------------------------------------------------
        # get shifted windows
        #-------------------------------------------------------------------------------
        #speckles move +Ax, use pixel border to allow slight Ay changes
        rect0 = (0, 32, 128,960)
        Ax,Ay,c, wins = misc.get_shifted_windows(imgs, rect0, ref_i)
        data.close()
        
        #%%-----------------------------------------------------------------------------
        # Fast
        #-------------------------------------------------------------------------------
        print 'Doing FAST'
        detector = detectors.FAST_Detector()
        res_fast = robustness_match(wins, detector, tolerance, ref_i)

        #%%-----------------------------------------------------------------------------
        # AGAST
        #-------------------------------------------------------------------------------
        print 'Doing AGAST'
        detector = detectors.AGAST_Detector()
        res_agast = robustness_match(wins, detector, tolerance, ref_i)

        #%%-----------------------------------------------------------------------------
        # Harris
        #-------------------------------------------------------------------------------
        print 'Doing Harris'

        detector = detectors.Harris_Detector()
        res_harris = robustness_match(wins, detector, tolerance, ref_i)

        #%%-----------------------------------------------------------------------------
        # USURF/DOG
        #-------------------------------------------------------------------------------
        print 'Doing USURF/DoG'
        detector = detectors.USURF_Detector()
        res_dog = robustness_match(wins, detector, tolerance, ref_i)

        #%%-----------------------------------------------------------------------------
        # SURF/oDOG
        #-------------------------------------------------------------------------------
        print 'Doing SURF/oDoG'
        detector = detectors.SURF_Detector()
        res_odog = robustness_match(wins, detector, tolerance, ref_i)

        #%%-----------------------------------------------------------------------------
        # ORB/oFAST
        #-------------------------------------------------------------------------------
        print 'Doing ORB/oFAST'
        detector = detectors.ORB_Detector()
        res_ofast = robustness_match(wins, detector, tolerance, ref_i)


        #%%-----------------------------------------------------------------------------
        # BRISK/oAGAST
        #-------------------------------------------------------------------------------
        print 'Doing BRISK/oAGAST'
        detector = detectors.BRISK_Detector()
        res_oagast = robustness_match(wins, detector, tolerance, ref_i)

        
        #%%-----------------------------------------------------------------------------
        # store processed results
        #-------------------------------------------------------------------------------
        surface_full = '{0} Rz={1} um'.format(surface, Rz)

        #store common items to the surface group
        common = {'Ax':Ax, 'Ay':Ay,'c':c, 'tolerance':tolerance }
        surface_group = lib.save_dict_to_hdf5( common , out, key=surface_full)

        #save methods
        lib.save_dict_to_hdf5(res_fast, surface_group, 'FAST')
        lib.save_dict_to_hdf5(res_agast, surface_group, 'AGAST')
        lib.save_dict_to_hdf5(res_harris, surface_group, 'Harris')
        lib.save_dict_to_hdf5(res_dog, surface_group, 'DoG')
        lib.save_dict_to_hdf5(res_odog, surface_group, 'ODoG')
        lib.save_dict_to_hdf5(res_ofast, surface_group, 'OFAST')
        lib.save_dict_to_hdf5(res_oagast, surface_group, 'OAGAST')
        
out.close()

#%%-----------------------------------------------------------------------------
# load from hdf5
#-------------------------------------------------------------------------------
#
###surface preparation method and roughtness:
#surf_treatment = 'flat lapping reaming'
##surf_treatment = 'grinding'
#Rz = 10  #10,6,3,1.6,1.0,0.55
#
##surf_treatment = 'horizontal milling'
##surf_treatment = 'vertical milling'
##surf_treatment = 'turning'
##Rz = 50 #50, 32,16,8 ,4,2.5
#
#filename = 'run2\\{0} Rz={1} um'.format(surf_treatment, Rz)
#print 'Loading data from :', filename
#
#data = h5py.File(filename, 'r')
#imgs = data['imgs']
#
#tolerance = 1.5
#ref_i = 0
##%%-----------------------------------------------------------------------------
## get shifted windows
##-------------------------------------------------------------------------------
##speckles move +Ax, use pixel border to allow slight Ay changes
#rect0 = (0, 32, 128,960)
#Ax,Ay,c, wins = lib.get_shifted_windows(imgs, rect0, ref_i)
#
##%%-----------------------------------------------------------------------------
## make example videos of windows
##-------------------------------------------------------------------------------
#import matplotlib.animation as manimation
#
##setup writer
#FFMpegWriter = manimation.writers['ffmpeg']
#metadata = dict(title='Translating window test', artist='Matplotlib',
#                comment='Centered windows from data set')
#writer = FFMpegWriter(fps=5, metadata=metadata)
#
##setup the figure to animate
#font = {'family' : 'sans',
#        'weight' : 'normal',
#        'size'   : 8}
#plt.rc('font', **font)
#
#fig = plt.figure(figsize=(6,3))
#ax1 = fig.add_axes( (0.05,0.1,0.8,0.8))
#
#im1 = ax1.imshow( wins[0,0:128,0:128], interpolation='nearest')
#ax1.set_xticks([])#hide labels
#ax1.set_yticks([])
#ax1.set_title('Centered windows')
#tx1 = ax1.text( 10,10, r'$A_x =0$', fontsize=12, color='w')
#
##animation loop
#with writer.saving(fig, "translating_window.mp4", dpi=600):
#    for n, win in enumerate(wins):
#        im1.set_data(win[0:128,0:128])
#        tx1.set_text(r'$A_x ={0:0.2f}$'.format(Ax[n]))
#        writer.grab_frame()
#
#
#

##%%-----------------------------------------------------------------------------
## plot vs C
##-------------------------------------------------------------------------------
#font = {'family' : 'sans',
#        'weight' : 'normal',
#        'size'   : 8}
#plt.rc('font', **font)
#
#cmap = plt.get_cmap('tab10')
#
#
#f = plt.figure('Robustness vs C dn={0}'.format(tolerance) )
#plt.plot(c, R_fast, '.', label='FAST', color=cmap(0))
#plt.plot(c, R_agast, '.', label='AGAST', color=cmap(1))
#plt.plot(c, R_harris, '.', label='Harris', color=cmap(2))
#plt.plot(c, R_dog, '.', label='DoG/USURF', color=cmap(3))
##plt.plot(c, R_odog, 'x', label='oDoG/SURF', color=cmap(3))
##plt.plot(c, R_ofast, 'x', label='oFAST/ORB', color=cmap(0))
##plt.plot(c, R_oagast, 'x', label='oAGAST/BRISK', color=cmap(1))
#
#plt.xlabel('Correlation Coefficent')
#plt.ylabel(r'Robustness $\frac{N_{matched}}{N_{0}} $')
#plt.legend()
# 
##%%-----------------------------------------------------------------------------
## plot vs shift
##-------------------------------------------------------------------------------
#font = {'family' : 'sans',
#        'weight' : 'normal',
#        'size'   : 8}
#plt.rc('font', **font)
#
#cmap = plt.get_cmap('tab10')
#
#f = plt.figure('Robustness vs shift')
#plt.plot(Ax, R_fast, '.', label='FAST', color=cmap(0))
#plt.plot(Ax, R_agast, '.', label='AGAST', color=cmap(1))
#plt.plot(Ax, R_harris, '.', label='Harris', color=cmap(2))
#plt.plot(Ax, R_dog, '.', label='DoG/USURF', color=cmap(3))
#plt.xlabel('Speckle translation, $A_x$')
#plt.ylabel(r'Robustness $\frac{N_{matched}}{N_{0}} $')
#plt.legend()
#
#
#%%-----------------------------------------------------------------------------
# plot vs N
#-------------------------------------------------------------------------------
#font = {'family' : 'sans',
#        'weight' : 'normal',
#        'size'   : 8}
#plt.rc('font', **font)
#
#f = plt.figure('Robustness vs N')
#plt.plot(N_fast, R_fast, '.', label='FAST')
#plt.plot(N_harris, R_harris, '.', label='Harris')
#plt.plot(N_dog, R_dog, '.', label='DoG')
#
#plt.xlabel('Number of features, N$_0$')
#plt.ylabel(r'Robustness $\frac{N_{matched}}{N_{0}} $')
#
#%%-----------------------------------------------------------------------------
# plot A vs N
#-------------------------------------------------------------------------------
#font = {'family' : 'sans',
#        'weight' : 'normal',
#        'size'   : 8}
#plt.rc('font', **font)
#
#f = plt.figure('Ax vs N')
#plt.plot(Ax, N_fast, '.', label='FAST')
#plt.plot(Ax, N_harris, '.', label='Harris')
#plt.plot(Ax, N_dog, '.', label='DoG')
#
#plt.xlabel('Speckle translation, $A_x$')
#plt.ylabel('Number of features, N$_0$')
#
#%%-----------------------------------------------------------------------------
# plot A vs c
#-------------------------------------------------------------------------------
#font = {'family' : 'sans',
#        'weight' : 'normal',
#        'size'   : 8}
#plt.rc('font', **font)
#
#f = plt.figure('Ax vs c')
#plt.plot(Ax, c, '.')
#
#plt.xlabel('Speckle translation, $A_x$')
#plt.ylabel('Correlation coefficent, c')
#
#
