"""
Examine orientation calculations.

Run sections as appropriate
"""

#%%-----------------------------------------------------------------------------
# setup
#-------------------------------------------------------------------------------
import lib.detectors as detectors
from lib import descriptors 
import lib.misc as misc
import lib.plotting as plotting
import h5py
import numpy as np
import cv2 as cv
import matplotlib.pyplot as plt
plt.ion()

surfaces = ( ('turning', (50,)),
            ('horizontal milling', (16,)),
            ('flat lapping reaming', (1.6,)),
            ('vertical milling', (16,)),)
            

def orientatation_comparision(  win0, win1, Dang, 
                                        detector, descriptor, matcher,
                                        radius, dn):
    """
    Compare the feature orientations between two windows.
    
    win0, win1  -   the two windows to compare
    Dang        -   the known angle change
    detector,descriptor, matcher    -   objects to be tested.
    radius      -   limit keypoint search to this radius to ensure all features 
                    are present even under rotation
    dn          -   tolerance in pixels for proximity search
    """
    #find keypoints in original window
    kps0 = detector.find_features(win0)

    #filter keypoints with radius greater than that given
    Cy,Cx  = np.array(win0.shape)//2 #center or rotation
    kps0 = misc.filter_kps( kps0, centre=(Cy,Cx), radius=radius)
    N = len(kps0)
    print 'Ref image Number of kps:',N

    #find keypoints in next window and filter keypoints with radius
    kps1 = detector.find_features(win1)
    kps1 = misc.filter_kps( kps1, centre=(Cy,Cx), radius=radius)
    print 'Image: ',n,'Dang: ',Dang, 'Number of kps:' ,len(kps1)

    #rotate the keypoints to reference frame
    #+ve here due to definition of Dang and image coords
    kps1_ref = misc.rotate_kps( kps1, Dang, centre=(Cy,Cx))

    #Find the number of kps that can be proximity matched
    matches = misc.proximity_match( kps0, kps1_ref, dn)
    kps0_v = [kps0[i] for i,j in matches]
    kps1_v = [kps1[j] for i,j in matches]
    N0 = len(matches)
    print 'Proximity matched: ', N0

    #find difference in calculated orientation angles for all keypoints that were proximity matched
    Dorient = [ kps1_ref[j].angle - kps0[i].angle for i,j in matches]
    Dorient = (np.array(Dorient)-180)%360 - 180 #wrap to +-180
    #print Dorient, kps1_ref[0].angle, kps0[0].angle
    #Dorient = abs(Dorient) #make magnitude
    print 'Change in orientation: ', Dorient.mean(), Dorient.std()

    #match these two sets by descriptors
    kps0_d, Ds0 = descriptor.compute_descriptors(win0, kps0_v)
    kps1_d, Ds1 = descriptor.compute_descriptors(win1, kps1_v)
    ms = matcher.match(Ds1, Ds0)
    print 'Descriptor matched: ', len(ms) 

    #orientation differences for sucessful matches and failures (bad matched, no matches)
    Do_matched_good = []    #matched and good
    Do_matched_bad  = []     #matched and bad
    i_failed = range(0,len(kps0_d))
    for Dmatch in ms:
        kp0 = kps0_d[ Dmatch.trainIdx]
        kp1 = kps1_d[ Dmatch.queryIdx]
        Do = ((kp1.angle +Dang -kp0.angle -180)%360 -180)
        i_failed.remove(Dmatch.trainIdx)
        if Dmatch.trainIdx == Dmatch.queryIdx:
            Do_matched_good.append(Do)
        else:
            Do_matched_bad.append(Do)
            
    Do_matched_good = np.array(Do_matched_good)
    Do_matched_bad = np.array(Do_matched_bad)
    Do_failed = np.array(Dorient[i_failed]) #not matched at all!
    
    #find R = N(<25deg) / N0
    N_orient = len( Dorient[abs(Dorient)<25.0])
    R = float(N_orient)/N0
    print 'N <25degree ', N_orient
    print 'R <25degree ', R

    return Do_matched_good, Do_matched_bad, Do_failed, N0, N_orient, R
    
    
#Need to use orientated detector for this!
orb_detector = detectors.ORB_Detector()      #oFAST
brisk_detector = detectors.BRISK_Detector()     #oFAST
surf_detector = detectors.SURF_Detector()     #oDoG
usurf_detector = detectors.USURF_Detector()     #DoG


usurf1 = descriptors.SURF( upright=True, extended=False)
usurf2 = descriptors.SURF( upright=True, extended=True)
surf1 = descriptors.SURF( upright=False, extended=False)
surf2 = descriptors.SURF( upright=False, extended=True)
orb16 = descriptors.ORB(patchSize=16)
orb32 = descriptors.ORB(patchSize=32)
orb48 = descriptors.ORB(patchSize=48)
orb64 = descriptors.ORB(patchSize=64)
brisk16 = descriptors.BRISK(patchSize=16)
brisk32 = descriptors.BRISK(patchSize=32)
brisk48 = descriptors.BRISK(patchSize=48)
brisk64 = descriptors.BRISK(patchSize=64)

#matchers to use
L1_norm = cv.BFMatcher(cv.NORM_L1, crossCheck=True)
L2_norm = cv.BFMatcher(cv.NORM_L2, crossCheck=True)
Hamming = cv.BFMatcher(cv.NORM_HAMMING, crossCheck=True)

to_test = ( (usurf_detector, usurf1, L1_norm),
            (surf_detector, surf1, L1_norm),   
            (orb_detector, orb32, Hamming),
            (brisk_detector, brisk32, Hamming), 
            )   
#%%-----------------------------------------------------------------------------
# 1) generate windows rotating data
#-------------------------------------------------------------------------------
tolerance = 1.5 #tolerance used in proximity matching
ref_i = 0                   #reference to this image
radius = 360           #size of windows to extract
pad = 32                #size of border padding


surface, Rzs = surfaces[2]
Rz = Rzs[0]
path = '..\\data\\dataset5 surface samples rotation\\'
filename = '{0} Rz={1} um.hdf5'.format(surface, Rz)

print 'Loading data from :', path+filename

data = h5py.File(path+filename, 'r')
imgs = data['imgs']
angs = data['angs'][:]
centre=(490,625) 
Ax, Ay, Dangs, c, wins, wins_ = misc.get_rotated_windows( imgs, angs, centre=centre, radius=radius, pad=pad, ref_i=ref_i, mask=False)

#%%-----------------------------------------------------------------------------
# 1) generate windows translation data
#-------------------------------------------------------------------------------
max_i = 50  #limit to valid shifts Ax

surface, Rzs = surfaces[1]
Rz = Rzs[0]

path = '..\\data\\dataset4 surface samples translation\\'
filename = '{0} Rz={1} um.hdf5'.format(surface, Rz)
data = h5py.File(path+filename, 'r')
imgs = data['imgs']
rect0 = (0, 32, 256,960)
Ax,Ay,c, wins = misc.get_shifted_windows(imgs, rect0, ref_i, max_i)
Dangs = np.zeros( wins.shape[0])

#%%-----------------------------------------------------------------------------
# 2) process single method
#-------------------------------------------------------------------------------
detector,descriptor, matcher = to_test[1]

#reference window
win0 = wins[ref_i]

#get window and angle delta
n=5  #5 = 100px, 2=10degree
Dang = Dangs[n]
win1 = wins[n]

Do_matched_good, Do_matched_bad, Do_failed, N0, Norient, R = orientatation_comparision(  win0, win1, Dang, detector, descriptor, matcher, radius, tolerance)
print 'Matches <25deg / good/bad/failed :', Norient, len(Do_matched_good), len(Do_matched_bad), len(Do_failed)

Do_mean_good = Do_matched_good.mean()
Do_mean_bad = Do_matched_bad.mean()
Do_mean_failed = Do_failed.mean()
print 'Mean absolute orientatation difference: good/bad/failed', Do_mean_good, Do_mean_bad,Do_mean_failed

Do_std_good = Do_matched_good.std()
Do_std_bad = Do_matched_bad.std()
Do_std_failed = Do_failed.std()
print 'Mean orientatation difference std dev: success/failed', Do_std_good, Do_std_bad, Do_std_failed


plt.figure()
plt.plot(Do_matched_bad, 'xr', ms=3, label='Bad matches')
plt.plot(Do_failed     , 'xb',ms=3, label='Failed matches')
plt.plot(Do_matched_good, '.g',ms=3, label='Good matches')


#%%-----------------------------------------------------------------------------
# process and combine all data...
#-------------------------------------------------------------------------------
out_path = '..\\results\\5.5.1 orientation_histogram.hdf5'
out = h5py.File(out_path, 'a')  

for detector,descriptor, matcher in to_test:

    #reference window
    win0 = wins[ref_i]

    good = []
    bad  = []
    failed = []
    for n in range(0, wins.shape[0]):
        win1 = wins[n]
        Dang = Dangs[n]
        
        Do_matched_good, Do_matched_bad, Do_failed, N0, Norient, R = orientatation_comparision(  win0, win1, Dang, detector, descriptor, matcher, radius, tolerance)

        good.append(Do_matched_good)
        bad.append(Do_matched_bad)
        failed.append(Do_failed)
        
    good = np.hstack(good)
    bad = np.hstack(bad)
    failed = np.hstack(failed)
    ugly = np.hstack((bad,failed)) #for stacking on histogram

    res = {'good': good, 'bad': bad, 'failed': failed, 'ugly': ugly}
    g = misc.save_dict_to_hdf5( res , out, key=descriptor.method)
    
out.close()

#%%-----------------------------------------------------------------------------
# figure histogram single method
#-------------------------------------------------------------------------------
data_path = '..\\results\\5.5.1 orientation_histogram.hdf5'
data = h5py.File(data_path, 'a')  

#select method
detector,descriptor, matcher = to_test[1]
res  = data[ descriptor.method ]

good  = res['good']
bad   = res['bad']
failed = res['failed']
ugly = res['ugly'] #np.hstack((bad,failed))


bins = np.linspace(-180,180,90)

font = {'family' : 'sans',
        'weight' : 'normal',
        'size'   : 8}
plt.rc('font', **font)
f = plt.figure(figsize=(3.25,2))

ax1 = f.add_axes( (0.2,0.25,0.7,0.7)) 

ax1.hist(ugly, bins, color='red', label='Bad')
ax1.hist(failed, bins, color='orange', label='Failed')
ax1.hist(good, bins, color='green', label='Good')
ax1.set_ylim(0,2500)
ax1.set_xticks((-180,-90,0,90,180))

plt.ylabel('Number\nof features')
plt.xlabel('Difference in calculated\nfeature orientation [$^o$]')

l =  ax1.legend(loc=(0.1,0.7), labelspacing=0.1, frameon=False, handletextpad=0.1, borderpad=0 )
data.close()