"""
Investigate feature descriptor orientation "robustness"
"""
#%%-----------------------------------------------------------------------------
# setup
#-------------------------------------------------------------------------------
import lib.detectors as detectors
from lib import descriptors 
import lib.misc as misc
import h5py
import numpy as np
import cv2 as cv

def orientation_robustness_translation(detector, descriptor, matcher,  wins, radius=100, ref_i=0, dn=1.5):
    """
    Test the orientation robustness of the detector given using translation 
    corrected rotating windows.
    
    1) Find features in reference window.
    2) For each new window:
        find features
        proximity match to reference set = Nv
        compare orientations angles of features and select those <25degrees 
    3)  Compute R = N<25 / N0
    """
    #reference window
    win0 = wins[ref_i]
    
    #find keypoints in original window
    kps0 = detector.find_features(win0)
    
    #filter keypoints with radius greater than that given
    Cy,Cx  = np.array(win0.shape)//2 #center or rotation
    kps0 = misc.filter_kps( kps0, centre=(Cy,Cx), radius=radius)
    N0 = len(kps0)
    print 'N0 : ',N0
    
    #Loop over all windows and match to first window found
    res = {}
    res['method'] = detector.method
    N = []      #Number of features found
    Nv = []     #Number of features matched by proximity
    Nm = []     #Number matched by descriptor
    No25 = []   #Number with orientation calculated to within 25degrees
    for n in range(0,len(wins) ):
        
        #get window and angle delta
        win1 = wins[n]
        
        #find keypoints in next window and filter keypoints with radius
        kps1 = detector.find_features(wins[n])
        kps1 = misc.filter_kps( kps1, centre=(Cy,Cx), radius=radius)
        N.append(len(kps1))
        print 'Image: ',n, 'Number of kps:' ,len(kps1), 

        #Find the number of kps that can be proximity matched
        matches = misc.proximity_match( kps0, kps1, dn)
        Nv.append( len(matches) )
        kps0_v = [kps0[i] for i,j in matches]
        kps1_v = [kps1[j] for i,j in matches]
        print 'Proximity matched: ', Nv[-1], 
        
        #find difference in calculated orientation angles for all keypoints
        Dorient = [ kps1[j].angle - kps0[i].angle for i,j in matches]
        Dorient = (np.array(Dorient)-180)%360 - 180 #wrap to +-180
        
        #find R = N(<25deg) / Nv
        N_orient = len(Dorient[abs(Dorient)<25.0])
        No25.append(N_orient)
        print 'No25 ', No25[-1],
        
        try:
            print 'R  ',float(N_orient)/Nv[-1]
        except:
            pass
    #convert to array types
    N = np.array(N, np.float64)       #number of features
    Nv = np.array(Nv, np.float64)     #number proximity matched
    Nm = np.array(Nm, np.float64)     #number matched by descriptor
    No25 = np.array(No25, np.float64)     #number with orientation difference <25degrees
    R = No25/Nv
    
    #store arrays to results
    res['N'] = N
    res['Nv'] = Nv
    res['Nm'] = Nm
    res['No25'] = No25
    res['R'] = R
    
    return res
    

#%%-----------------------------------------------------------------------------
# setup samples names etc
#-------------------------------------------------------------------------------
tolerance = 1.5 #tolerance used in proximity matching
ref_i = 0                   #reference to this image
rect0 = (0, 32, 256 ,960)   #window size to use
max_i = 50  #limit to valid shifts Ax

#surface preparation methods to process
surfaces = ( ('horizontal milling', (2.5,16,50)),
            ('flat lapping reaming', (0.55,1.6,10)),
            ('vertical milling', (2.5,16,50)),
            ('turning', (2.5,16,50)),
            ('grinding', (0.55,1.6,10))  )
            
surfaces = ( ('turning', (2.5,16,50)),
            ('grinding', (0.55,1.6,10))  )
                      
#surfaces = ( ('turning', (50,)),
#            ('horizontal milling', (16,)),
#            ('flat lapping reaming', (1.6,)),
#            ('vertical milling', (16,)),)
#            
path = '..\\data\\dataset4 surface samples translation\\'
out_path = '..\\results\\5.5.1 orientation_translation.hdf5'


#Need to use orientated detector for this!
orb_detector = detectors.ORB_Detector()      #oFAST
brisk_detector = detectors.BRISK_Detector()     #oFAST
surf_detector = detectors.SURF_Detector()     #oDoG

usurf1 = descriptors.SURF( upright=True, extended=False)
usurf2 = descriptors.SURF( upright=True, extended=True)
surf1 = descriptors.SURF( upright=False, extended=False)
surf2 = descriptors.SURF( upright=False, extended=True)
orb16 = descriptors.ORB(patchSize=16)
orb32 = descriptors.ORB(patchSize=32)
orb48 = descriptors.ORB(patchSize=48)
orb64 = descriptors.ORB(patchSize=64)
brisk16 = descriptors.BRISK(patchSize=16)
brisk32 = descriptors.BRISK(patchSize=32)
brisk48 = descriptors.BRISK(patchSize=48)
brisk64 = descriptors.BRISK(patchSize=64)

#matchers to use
L1_norm = cv.BFMatcher(cv.NORM_L1, crossCheck=True)
L2_norm = cv.BFMatcher(cv.NORM_L2, crossCheck=True)
Hamming = cv.BFMatcher(cv.NORM_HAMMING, crossCheck=True)

to_test = ( (surf_detector, surf1, L1_norm),   
            (orb_detector, orb32, Hamming),
            (brisk_detector, brisk32, Hamming), )

#%%-----------------------------------------------------------------------------
# Loop over each sample
#-------------------------------------------------------------------------------
out = h5py.File(out_path, 'a')  

for surface, Rzs in surfaces:
    for Rz in Rzs:
        print '------------------------------ {0} {1} ------------------------------'.format(surface, Rz)

        #-------------------------------------------------------------------------------
        # load from hdf5
        #-------------------------------------------------------------------------------
        filename = '{0} Rz={1} um.hdf5'.format(surface, Rz)

        print 'Loading data from :', path+filename

        data = h5py.File(path+filename, 'r')
        imgs = data['imgs']

        #-------------------------------------------------------------------------------
        # get rotated/centered windows
        #-------------------------------------------------------------------------------
        Ax,Ay,c, wins = misc.get_shifted_windows(imgs, rect0, ref_i, max_i)
        data.close()
        
        #-------------------------------------------------------------------------------
        # set up group to store processed results
        #-------------------------------------------------------------------------------
        surface_full = '{0} Rz={1} um'.format(surface, Rz)
        #store common items to the surface group
        common = {'Ax':Ax, 'Ay':Ay, 'c':c, 'tolerance':tolerance, 'rect0':rect0 }
        surface_group = misc.save_dict_to_hdf5( common , out, key=surface_full)

        #-------------------------------------------------------------------------------
        # run robustness test
        #-------------------------------------------------------------------------------
        for detector, descriptor, matcher  in to_test:
            
            print 'Doing', detector.method
            res = orientation_robustness_translation(detector, descriptor, matcher,
                                                   wins, ref_i=ref_i, dn=tolerance)
            #save to file
            misc.save_dict_to_hdf5(res, surface_group, res['method'])      
        
out.close()