"""
Investigate feature descriptor "robustness" - the ability to describe the same 
features/speckles in multiple frames with changing/decorrelation

Method/robustness in the presence of translations:
    1) find the shift between frames using normalised cross-correaltion
    2) Find keypoints in first image window defined by rect0. 
    3) Find keypoints in a shifted window in the next image, where the window is
    shifted to account for the speckle shift and ensure that all speckles are 
    still present and only interframe changes/decorrelation of the speckles 
    influences the detection
    
    Then for robustness:
    
    4) Find matching keypoints using brute force proximity matching.
    5) Calculate and compare the descriptors using different methods

"""
#%%-----------------------------------------------------------------------------
# setup
#-------------------------------------------------------------------------------
import numpy as np
import matplotlib.pyplot as plt
plt.ion()
import time
import cv2 as cv

from lib import detectors
from lib import descriptors 
from lib import misc

import h5py

#%%-----------------------------------------------------------------------------
# function to do robustness check on rotating windows
#-------------------------------------------------------------------------------   
def rotation_robustness_match(  wins, Dangs, detector, descriptor, matcher, 
                                radius=100, ref_i=0, dn=1.5):
    """
    Do robustness check on rotating window data (see get_rotated_windows)
    
        *** Attempt to match all keypoints present then verify by proximity ***

    1) Find keypoints in first image window
    2) Find keypoints in a rotated window
    3) Find max number of matching keypoints using proximity test, Nv
    4) Compute descriptors and match
    5) Divide matches into verified (Nmv) and unverified (Nmu) matches based upon proximity
        and find the mean and standard deviation of their descriptor distances.
    6) Compute R = Nmv / Nv
    
    Args:
    radius  -   radius to search for features in
    dn      -   promitity test tolerance
    ref_i   -   reference frame index

    Returns results dictionary with keys: 
    method  -   method used
    N   -   Number of features found
    Nm  -   Number of features matched by descriptor
    Nv  -   Number of features matched by proximity
    Nmv -   Number of features matched by descriptor and verified by proximity
    Nmu -   Number of features matched by descriptor but not verified (bad matches?)

    dv_mean -   mean distance of verified matches
    du_mean -   mean distance of unverified matches
    dv_std  -   standard deviation of distance of verified matches
    du_std  -   standard deviation of distance of unverified matches
    
    R = Nmv/Nv - robustness
    
    """
    #reference window
    win0 = wins[ref_i]
    
    #find keypoints in original window
    kps0 = detector.find_features(win0)
    
    #filter keypoints with radius greater than that given
    Cy,Cx  = np.array(win0.shape)//2 #center or rotation
    kps0 = misc.filter_kps( kps0, centre=(Cy,Cx), radius=radius)
    N0 = len(kps0)
    print 'N0 : ',N0
    
    #compute descriptors for this reference set
    kps0_d, Ds0 = descriptor.compute_descriptors(win0, kps0)

    #Loop over all windows and match to first window found
    res = {}
    res['method'] = descriptor.method
    N = []      #Number of features found
    Nm = []     #Number of feature smatched by descriptor
    Nv = []     #Number of features matched by proximity
    Nmv = []    #Number of features matched by descriptor and verified by proximity
    Nmu = []    #Number of features matched by descriptor but not verified (bad matches?)

    dv_mean = []    #mean distance of verified matches
    du_mean = []    #mean distance of unverified (bad?) matches
    dv_std = []     #standard deviation of distance of verified matches
    du_std = []     #standard deviation of distance of unverified matches
    
    for n in range(0,len(wins) ):
        
        #get window and angle delta
        Dang = Dangs[n]
        win1 = wins[n]
        
        #find keypoints in next window and filter keypoints with radius
        kps1 = detector.find_features(wins[n])
        kps1 = misc.filter_kps( kps1, centre=(Cy,Cx), radius=radius)
        N.append(len(kps1))
        print 'Image: ',n,'Dang: ',Dang, 'Number of kps:' ,len(kps1), 

        #match the two sets by descriptors
        kps1_d, Ds1 = descriptor.compute_descriptors(win1, kps1)
        ms = matcher.match(Ds1, Ds0)
        Nm.append(len(ms))
        print 'Descriptor matched: ', Nm[-1], 
        
        #rotate the keypoints to reference frame (i.e. -Dang)
        #+ve here due to definition of Dang and image coords
        kps1_ref = misc.rotate_kps( kps1_d, Dang, centre=(Cy,Cx))

        #Find the number of kps that can be proximity matched
        matches = misc.proximity_match( kps0_d, kps1_ref, dn)
        Nv.append( len(matches) )
        print 'Proximity matched: ', Nv[-1], 
        
        #verifiy each descriptor match is within tolerance dn specified.
        d_verified = []
        d_unverified = []
        for Dmatch in ms:
            #pull out matching keypoints
            kp0 = kps0_d[Dmatch.trainIdx]
            kp1 = kps1_ref[Dmatch.queryIdx]   #use kp from rotated set!

            #check proximty distance and store descriptor difference
            pd = np.sqrt( (kp1.pt[0]-kp0.pt[0])**2 + (kp1.pt[1]-kp0.pt[1])**2 )
            if pd<=dn:
                d_verified.append(Dmatch.distance)
            else:
                d_unverified.append(Dmatch.distance)
            
        #store counts
        Nmv.append(  len(d_verified) )
        Nmu.append(  len(d_unverified) )
        
        #make arrays
        d_verified = np.array(d_verified)
        d_unverified = np.array(d_unverified)
        print 'verified: ', Nmv[-1],
        
        #calculate the stats and store results
        dv_mean.append(d_verified.mean())
        dv_std.append(d_verified.std())
        du_mean.append(d_unverified.mean())
        du_std.append(d_unverified.std())

        print 'dv mean: {0:0.2f} std: {1:0.2f}'.format( dv_mean[-1], dv_std[-1]),
        print 'du mean: {0:0.2f} std: {1:0.2f}'.format( du_mean[-1], du_std[-1])

    #convert to array types
    N = np.array(N, np.float64)       #number of features
    Nv = np.array(Nv, np.float64)     #number proximity matched
    Nm = np.array(Nm, np.float64)     #number of descriptor matched features
    Nmv = np.array(Nmv, np.float64)   #number of verified (i.e. also proximity matched) descriptor matched features
    Nmu = np.array(Nmv, np.float64)   #number of unverified (i.e. not also proximity matched) descriptor matched features

    #distances
    dv_mean = np.array(dv_mean)       #mean distance of verified matches
    du_mean = np.array(du_mean)       #mean distance of unverified matches
    dv_std = np.array(dv_std)         #std. dev. of distance of verified matches
    du_std = np.array(du_std)         #std. dev. of distance of unverified matches
    
    #store arrays to results
    res['N'] = N
    res['Nv'] = Nv
    res['Nm'] = Nm    
    res['Nmv'] = Nmv
    res['Nmu'] = Nmu
    res['dv_mean'] = dv_mean
    res['du_mean'] = du_mean
    res['dv_std'] = dv_std
    res['du_std'] = du_std
    
    #calculate robustness
    res['R'] = Nmv/Nv                  #robustness
    
    return res

def rotation_robustness_match2(  wins, Dangs, detector, descriptor, matcher, 
                                radius=100, ref_i=0, dn=1.5):
    """
    Do robustness check on rotating window data (see get_rotated_windows)
    
        *** Attempt to match only keypoints verified by proximity ***

    1) Find keypoints in first image window
    2) Find keypoints in a rotated window
    3) Find max number of matching keypoints using proximity test, Nv
    4) Compute descriptors and match
    5) Divide matches into verified (Nmv) and unverified (Nmu) matches based upon proximity
        and find the mean and standard deviation of their descriptor distances.
    6) Compute R = Nmv / Nv
    
    Args:
    radius  -   radius to search for features in
    dn      -   promitity test tolerance
    ref_i   -   reference frame index

    Returns results dictionary with keys: 
    method  -   method used
    N   -   Number of features found
    Nm  -   Number of features matched by descriptor
    Nv  -   Number of features matched by proximity
    Nmv -   Number of features matched by descriptor and verified by proximity
    Nmu -   Number of features matched by descriptor but not verified (bad matches?)

    dv_mean -   mean distance of verified matches
    du_mean -   mean distance of unverified matches
    dv_std  -   standard deviation of distance of verified matches
    du_std  -   standard deviation of distance of unverified matches
    
    R = Nmv/Nv - robustness
    
    """
    #reference window
    win0 = wins[ref_i]
    
    #find keypoints in original window
    kps0 = detector.find_features(win0)
    
    #filter keypoints with radius greater than that given
    Cy,Cx  = np.array(win0.shape)//2 #center or rotation
    kps0 = misc.filter_kps( kps0, centre=(Cy,Cx), radius=radius)
    N0 = len(kps0)
    print 'N0 : ',N0
    
    #Loop over all windows and match to first window found
    res = {}
    res['method'] = descriptor.method
    N = []      #Number of features found
    Nm = []     #Number of feature smatched by descriptor
    Nv = []     #Number of features matched by proximity
    Nmv = []    #Number of features matched by descriptor and verified by proximity
    Nmu = []    #Number of features matched by descriptor but not verified (bad matches?)

    dv_mean = []    #mean distance of verified matches
    du_mean = []    #mean distance of unverified (bad?) matches
    dv_std = []     #standard deviation of distance of verified matches
    du_std = []     #standard deviation of distance of unverified matches
    
    for n in range(0,len(wins) ):
        
        #get window and angle delta
        Dang = Dangs[n]
        win1 = wins[n]
        
        #find keypoints in next window and filter keypoints with radius
        kps1 = detector.find_features(wins[n])
        kps1 = misc.filter_kps( kps1, centre=(Cy,Cx), radius=radius)
        N.append(len(kps1))
        print 'Image: ',n,'Dang: ',Dang, 'Number of kps:' ,len(kps1), 

        #rotate the keypoints to reference frame
        #+ve here due to definition of Dang and image coords
        kps1_ref = misc.rotate_kps( kps1, Dang, centre=(Cy,Cx))

        #Find the number of kps that can be proximity matched
        matches = misc.proximity_match( kps0, kps1_ref, dn)
        Nv.append( len(matches) )
        kps0_v = [kps0[i] for i,j in matches]
        kps1_v = [kps1[j] for i,j in matches]
        print 'Proximity matched: ', Nv[-1], 

        #match these two sets by descriptors
        kps0_d, Ds0 = descriptor.compute_descriptors(win0, kps0_v)
        kps1_d, Ds1 = descriptor.compute_descriptors(win1, kps1_v)
        ms = matcher.match(Ds1, Ds0)
        Nm.append(len(ms))
        print 'Descriptor matched: ', Nm[-1], 
        
        #Check for valid matches
        #indices should be same!
        d_verified = []
        d_unverified = []
        for Dmatch in ms:
            if Dmatch.trainIdx == Dmatch.queryIdx:
                d_verified.append(Dmatch.distance)
            else:
                d_unverified.append(Dmatch.distance)
            
        #store counts
        Nmv.append(  len(d_verified) )
        Nmu.append(  len(d_unverified) )
        
        #make arrays
        d_verified = np.array(d_verified)
        d_unverified = np.array(d_unverified)
        print 'verified: ', Nmv[-1],
        
        #calculate the stats and store results
        dv_mean.append(d_verified.mean())
        dv_std.append(d_verified.std())
        du_mean.append(d_unverified.mean())
        du_std.append(d_unverified.std())

        print 'dv mean: {0:0.2f} std: {1:0.2f}'.format( dv_mean[-1], dv_std[-1]),
        print 'du mean: {0:0.2f} std: {1:0.2f}'.format( du_mean[-1], du_std[-1])

    #convert to array types
    N = np.array(N, np.float64)       #number of features
    Nv = np.array(Nv, np.float64)     #number proximity matched
    Nm = np.array(Nm, np.float64)     #number of descriptor matched features
    Nmv = np.array(Nmv, np.float64)   #number of verified (i.e. also proximity matched) descriptor matched features
    Nmu = np.array(Nmv, np.float64)   #number of unverified (i.e. not also proximity matched) descriptor matched features

    #distances
    dv_mean = np.array(dv_mean)       #mean distance of verified matches
    du_mean = np.array(du_mean)       #mean distance of unverified matches
    dv_std = np.array(dv_std)         #std. dev. of distance of verified matches
    du_std = np.array(du_std)         #std. dev. of distance of unverified matches
    
    #store arrays to results
    res['N'] = N
    res['Nv'] = Nv
    res['Nm'] = Nm    
    res['Nmv'] = Nmv
    res['Nmu'] = Nmu
    res['dv_mean'] = dv_mean
    res['du_mean'] = du_mean
    res['dv_std'] = dv_std
    res['du_std'] = du_std
    
    #calculate robustness
    res['R'] = Nmv/Nv                  #robustness
    
    return res


#Paper method!
def rotation_robustness_match3(  wins, Dangs, detector, descriptor, matcher, 
                                radius=100, ref_i=0, dn=1.5):
    """
    Do robustness check on rotating window data (see get_rotated_windows)
    
    *** Always calculate descriptors at keypoints found in first window and try to match
        this ensures the same points are examined as windows are translation corrected
        Orientation computation is also always correct!         ***
        
    1) Find keypoints in first image window
    2) rotate keypoints to find position in new window
    3) Compute descriptors for these keypoints in both and match
    4) Divide matches into verified (Nmv) and unverified (Nmu) matches
        and find the mean and standard deviation of their descriptor distances.
    5) Compute R = Nmv / Nv

    Args:
    dn      -   promitity test tolerance
    ref_i   -   reference frame index
    border  -   exclude features inside the width border region
                (some descriptors excluded those to close to image edge)

    Returns results dictionary with keys: 
    method  -   method used
    N   -   Number of features found
    Nm  -   Number of features matched by descriptor
    Nmv -   Number of features matched by descriptor and verified by proximity
    Nmu -   Number of features matched by descriptor but not verified (bad matches?)

    dv_mean -   mean distance of verified matches
    du_mean -   mean distance of unverified matches
    dv_std  -   standard deviation of distance of verified matches
    du_std  -   standard deviation of distance of unverified matches
    
    R = Nmv/N - robustness
    
    """
    #reference window
    win0 = wins[ref_i]
    
    #find keypoints in original window
    kps0 = detector.find_features(win0)
    
    #filter keypoints with radius greater than that given
    Cy,Cx  = np.array(win0.shape)//2 #center or rotation
    kps0 = misc.filter_kps( kps0, centre=(Cy,Cx), radius=radius)
    N0 = len(kps0)
    print 'N0 : ',N0
    
    #find reference descriptors
    kps0_d, Ds0 = descriptor.compute_descriptors(win0, kps0)

    #Loop over all windows and match to first window found
    res = {}
    res['method'] = descriptor.method
    Nm = []     #Number of feature smatched by descriptor
    Nmv = []    #Number of features matched by descriptor and verified by proximity
    Nmu = []    #Number of features matched by descriptor but not verified (bad matches?)

    dv_mean = []    #mean distance of verified matches
    du_mean = []    #mean distance of unverified (bad?) matches
    dv_std = []     #standard deviation of distance of verified matches
    du_std = []     #standard deviation of distance of unverified matches
    
    for n in range(0,len(wins) ):
        
        #get window and angle delta
        Dang = Dangs[n]
        win1 = wins[n]
        
        print 'Image: ',n,'Dang: ',Dang, 

        #rotate the keypoint set to positions in win1
        # from win0 to win1 is -Dang, center of window is Center of rotation
        kps1 = misc.rotate_kps( kps0, -Dang, centre=(Cy,Cx))
        
        #compute descriptors in new window
        win1 = wins[n]
        kps1_d, Ds1 = descriptor.compute_descriptors(win1, kps1)
        
        #match these two verified sets by descriptors
        ms = matcher.match(Ds1, Ds0)
        Nm.append(len(ms))
        print 'Descriptor matched: ', Nm[-1], 
        
        #verifiy each descriptor match (should have same indices)
        d_verified = []
        d_unverified = []
        for Dmatch in ms:
            if Dmatch.trainIdx==Dmatch.queryIdx:
                d_verified.append(Dmatch.distance )
            else:
                d_unverified.append(Dmatch.distance )

        #store counts
        Nmv.append(  len(d_verified) )
        Nmu.append(  len(d_unverified) )
        
        #make arrays
        d_verified = np.array(d_verified)
        d_unverified = np.array(d_unverified)
        print 'verified: ', Nmv[-1], 'unverified: ', Nmu[-1],

        #calculate the stats and store results
        dv_mean.append(d_verified.mean())
        dv_std.append(d_verified.std())
        du_mean.append(d_unverified.mean())
        du_std.append(d_unverified.std())

        print 'dv mean: {0:0.2f} std: {1:0.2f}'.format( dv_mean[-1], dv_std[-1]),
        print 'du mean: {0:0.2f} std: {1:0.2f}'.format( du_mean[-1], du_std[-1])

    #convert to array types
    Nm = np.array(Nm, np.float64)     #number of descriptor matched features
    Nmv = np.array(Nmv, np.float64)   #number of verified (i.e. also proximity matched) descriptor matched features
    Nmu = np.array(Nmu, np.float64)   #number of unverified (i.e. not also proximity matched) descriptor matched features

    #distances
    dv_mean = np.array(dv_mean)       #mean distance of verified matches
    du_mean = np.array(du_mean)       #mean distance of unverified matches
    dv_std = np.array(dv_std)         #std. dev. of distance of verified matches
    du_std = np.array(du_std)         #std. dev. of distance of unverified matches
    
    #store arrays to results
    res['N'] = N0
    res['Nm'] = Nm    
    res['Nmv'] = Nmv
    res['Nmu'] = Nmu
    res['dv_mean'] = dv_mean
    res['du_mean'] = du_mean
    res['dv_std'] = dv_std
    res['du_std'] = du_std
    
    #calculate robustness
    res['R'] = Nmv/N0                  #robustness
    
    return res
    
#%%-----------------------------------------------------------------------------
# setup samples names etc
#-------------------------------------------------------------------------------
tolerance = 1.5 #tolerance used in proximity matching
ref_i = 0                   #reference to this image
radius = 360           #size of windows to extract
pad = 32                #size of border padding

#surface preparation methods to process
surfaces = ( ('horizontal milling', (2.5,16,50)),
            ('flat lapping reaming', (0.55,1.6,10)),
            ('vertical milling', (2.5,16,50)),
            ('turning', (2.5,16,50)),
            ('grinding', (0.55,1.6,10))  )

#surfaces = ( ('turning', (50,)),
#            ('horizontal milling', (16,)),
#            ('flat lapping reaming', (1.6,)),
#            ('vertical milling', (16,)),)
            
path = '..\\data\\dataset5 surface samples rotation\\'

out_path = '..\\results\\5.5.2 descriptor_rotation_robustness.hdf5'

#%%-----------------------------------------------------------------------------
# set methods to test
#-------------------------------------------------------------------------------

#Need to use orientated detector for this!
orb_detector = detectors.ORB_Detector()      #oFAST
brisk_detector = detectors.BRISK_Detector()     #oFAST
surf_detector = detectors.SURF_Detector()     #oDoG
usurf_detector = detectors.USURF_Detector()     #DoG

usurf1 = descriptors.SURF( upright=True, extended=False)
surf1 = descriptors.SURF( upright=False, extended=False)
orb32 = descriptors.ORB(patchSize=32)
brisk32 = descriptors.BRISK(patchSize=32)

#matchers to use
L1_norm = cv.BFMatcher(cv.NORM_L1, crossCheck=True)
L2_norm = cv.BFMatcher(cv.NORM_L2, crossCheck=True)
Hamming = cv.BFMatcher(cv.NORM_HAMMING, crossCheck=True)

to_test = ( (usurf_detector, usurf1, L1_norm), 
            (surf_detector, surf1, L1_norm),   
            (orb_detector, orb32, Hamming),
            (brisk_detector, brisk32, Hamming), 
            )
    
#%%-----------------------------------------------------------------------------
# Loop over each sample
#-------------------------------------------------------------------------------
out = h5py.File(out_path, 'a')  

for surface, Rzs in surfaces:
    for Rz in Rzs:
        print '------------------------------ {0} {1} ------------------------------'.format(surface, Rz)

        #%%-----------------------------------------------------------------------------
        # load from hdf5
        #-------------------------------------------------------------------------------
        filename = '{0} Rz={1} um.hdf5'.format(surface, Rz)

        print 'Loading data from :', path+filename

        data = h5py.File(path+filename, 'r')
        imgs = data['imgs']
        angs = data['angs'][:]

        #%%-----------------------------------------------------------------------------
        # get rotated/centered windows
        #-------------------------------------------------------------------------------
        centre=(490,625) 
        Ax, Ay, Dangs, c, wins, wins_ = misc.get_rotated_windows( imgs, angs, centre=centre, radius=radius, pad=pad, ref_i=ref_i, mask=False)
        #wins are centered window but still rotated originals
        #wins_ are centered and rotated to same orientation for testing comparision
        
        #-------------------------------------------------------------------------------
        # set up group to store processed results
        #-------------------------------------------------------------------------------
        surface_full = '{0} Rz={1} um'.format(surface, Rz)
        #store common items to the surface group
        common = {'Ax':Ax, 'Ay':Ay,'Dangs':Dangs, 'c':c, 'tolerance':tolerance, 'radius':radius }
        surface_group = misc.save_dict_to_hdf5( common , out, key=surface_full)

        #-------------------------------------------------------------------------------
        # run robustness test
        #-------------------------------------------------------------------------------
        for detector, descriptor, matcher in to_test:
            
            print 'Doing ', descriptor.method
            
            res = rotation_robustness_match3( wins, Dangs, detector, descriptor, 
                                                matcher, radius=radius, 
                                                ref_i=ref_i, dn=tolerance)
                                
            #save to file
            misc.save_dict_to_hdf5(res, surface_group, res['method'])      
        
out.close()