"""
Investigate feature descriptor "robustness" - the ability to describe the same 
features/speckles in multiple frames with changing/decorrelation

Method/robustness in the presence of translations:
    1) find the shift between frames using normalised cross-correaltion
    2) Find keypoints in first image window defined by rect0. 
    3) Find keypoints in a shifted window in the next image, where the window is
    shifted to account for the speckle shift and ensure that all speckles are 
    still present and only interframe changes/decorrelation of the speckles 
    influences the detection
    
    Then for robustness:
    
    4) Find matching keypoints using brute force proximity matching.
    5) Calculate and compare the descriptors using different methods

"""
#%%-----------------------------------------------------------------------------
# setup
#-------------------------------------------------------------------------------
import numpy as np
import matplotlib.pyplot as plt
plt.ion()
import time
import cv2 as cv

from lib import detectors
from lib import descriptors
from lib import misc

import h5py

#%%-----------------------------------------------------------------------------
# function to do robustness check on setup of translating windows
#-------------------------------------------------------------------------------   
def robustness_match( wins, detector, descriptor, matcher,  dn=1.5, ref_i=0, border=48):
    """
    Do robustness check on shifted windows (see get_shifted_windows).
    
    *** Attempt to match all keypoints present then verify by proximity ***
    
    1) Find keypoints in first image window
    2) Find keypoints in a shifted window
    3) Find max number of matching keypoints using proximity test, Nv
    4) Compute descriptors and match
    5) Divide matches into verified (Nmv) and unverified (Nmu) matches based upon proximity
        and find the mean and standard deviation of their descriptor distances.
    6) Compute R = Nmv / Nv

    Args:
    dn      -   promitity test tolerance
    ref_i   -   reference frame index
    border  -   exclude features inside the width border region
                (some descriptors excluded those to close to image edge)

    Returns results dictionary with keys: 
    method  -   method used
    N   -   Number of features found
    Nm  -   Number of features matched by descriptor
    Nv  -   Number of features matched by proximity
    Nmv -   Number of features matched by descriptor and verified by proximity
    Nmu -   Number of features matched by descriptor but not verified (bad matches?)

    dv_mean -   mean distance of verified matches
    du_mean -   mean distance of unverified matches
    dv_std  -   standard deviation of distance of verified matches
    du_std  -   standard deviation of distance of unverified matches
    
    R = Nmv/Nv - robustness
    """

    #find keypoints in original window v
    win0 = wins[ref_i]
    kps0 = detector.find_features(win0) 
    kps0 = misc.filter_kps_border( kps0, win0.shape, border)
    N0 = len(kps0)
    
    #compute descriptors for this reference set
    kps0_d, Ds0 = descriptor.compute_descriptors(win0, kps0)

    #Loop over all windows and match to first window found
    res = {}
    res['method'] = descriptor.method
    N = []      #Number of features found
    Nm = []     #Number of feature smatched by descriptor
    Nv = []     #Number of features matched by proximity
    Nmv = []    #Number of features matched by descriptor and verified by proximity
    Nmu = []    #Number of features matched by descriptor but not verified (bad matches?)

    dv_mean = []    #mean distance of verified matches
    du_mean = []    #mean distance of unverified (bad?) matches
    dv_std = []     #standard deviation of distance of verified matches
    du_std = []     #standard deviation of distance of unverified matches
    
    for n in range(0,len(wins) ):
        
        #find features in next window
        win1 = wins[n]
        kps1 = detector.find_features(win1)
        kps1 = misc.filter_kps_border( kps1, win1.shape, border)
        print 'Image: ',n, 'Number of kps:' , len(kps1), 
        N.append(len(kps1))
        
        #match the two sets by descriptors
        kps1_d, Ds1 = descriptor.compute_descriptors(win1, kps1)
        ms = matcher.match(Ds1, Ds0)
        Nm.append(len(ms))
        print 'Descriptor matched: ', Nm[-1], 
        
        #match by proximity (to ensure same keypoints in both, used as max possible matches)
        matches = misc.proximity_match( kps0_d, kps1_d, dn)
        Nv.append( len(matches) )
        print 'Proximity matched: ', Nv[-1], 

        #verifiy each descriptor match is within tolerance dn specified.
        d_verified = []
        d_unverified = []
        for Dmatch in ms:
            #pull out matching keypoints
            kp0 = kps0_d[Dmatch.trainIdx]
            kp1 = kps1_d[Dmatch.queryIdx]

            #check proximity distance and store descriptor distance
            pd = np.sqrt( (kp1.pt[0]-kp0.pt[0])**2 + (kp1.pt[1]-kp0.pt[1])**2 )
            if pd<=dn:
                d_verified.append(Dmatch.distance )
            else:
                d_unverified.append(Dmatch.distance )

        #store counts
        Nmv.append(  len(d_verified) )
        Nmu.append(  len(d_unverified) )
        
        #make arrays
        d_verified = np.array(d_verified)
        d_unverified = np.array(d_unverified)
        print 'verified: ', Nmv[-1], 'unverified: ', Nmu[-1],

        #calculate the stats and store results
        dv_mean.append(d_verified.mean())
        dv_std.append(d_verified.std())
        du_mean.append(d_unverified.mean())
        du_std.append(d_unverified.std())

        print 'dv mean: {0:0.2f} std: {1:0.2f}'.format( dv_mean[-1], dv_std[-1]),
        print 'du mean: {0:0.2f} std: {1:0.2f}'.format( du_mean[-1], du_std[-1])

    #convert to array types
    N = np.array(N, np.float64)       #number of features
    Nv = np.array(Nv, np.float64)     #number proximity matched
    Nm = np.array(Nm, np.float64)     #number of descriptor matched features
    Nmv = np.array(Nmv, np.float64)   #number of verified (i.e. also proximity matched) descriptor matched features
    Nmu = np.array(Nmu, np.float64)   #number of unverified (i.e. not also proximity matched) descriptor matched features

    #distances
    dv_mean = np.array(dv_mean)       #mean distance of verified matches
    du_mean = np.array(du_mean)       #mean distance of unverified matches
    dv_std = np.array(dv_std)         #std. dev. of distance of verified matches
    du_std = np.array(du_std)         #std. dev. of distance of unverified matches
    
    #store arrays to results
    res['N'] = N
    res['Nv'] = Nv
    res['Nm'] = Nm    
    res['Nmv'] = Nmv
    res['Nmu'] = Nmu
    res['dv_mean'] = dv_mean
    res['du_mean'] = du_mean
    res['dv_std'] = dv_std
    res['du_std'] = du_std
    
    #calculate robustness
    res['R'] = Nmv/Nv                  #robustness
    
    return res

def robustness_match2( wins, detector, descriptor, matcher,  dn=1.5, ref_i=0, border=48):
    """
    Do robustness check on shifted windows (see get_shifted_windows).
    
    *** Only try to match verified keypoints that are present in both ***
    
    1) Find keypoints in first image window
    2) Find keypoints in a shifted window
    3) Find matching keypoints using proximity test, Nv
    4) Compute descriptors for keypoint present in both and match
    5) Divide matches into verified (Nmv) and unverified (Nmu) matches based upon proximity
        and find the mean and standard deviation of their descriptor distances.
    6) Compute R = Nmv / Nv

    Args:
    dn      -   promitity test tolerance
    ref_i   -   reference frame index
    border  -   exclude features inside the width border region
                (some descriptors excluded those to close to image edge)

    Returns results dictionary with keys: 
    method  -   method used
    N   -   Number of features found
    Nm  -   Number of features matched by descriptor
    Nv  -   Number of features matched by proximity
    Nmv -   Number of features matched by descriptor and verified by proximity
    Nmu -   Number of features matched by descriptor but not verified (bad matches?)

    dv_mean -   mean distance of verified matches
    du_mean -   mean distance of unverified matches
    dv_std  -   standard deviation of distance of verified matches
    du_std  -   standard deviation of distance of unverified matches
    
    R = Nmv/Nv - robustness
    """

    #find keypoints in original window v
    win0 = wins[ref_i]
    kps0 = detector.find_features(win0) 
    kps0 = misc.filter_kps_border( kps0, win0.shape, border)
    N0 = len(kps0)
    
    #compute descriptors for this reference set

    #Loop over all windows and match to first window found
    res = {}
    res['method'] = descriptor.method
    N = []      #Number of features found
    Nm = []     #Number of feature smatched by descriptor
    Nv = []     #Number of features matched by proximity
    Nmv = []    #Number of features matched by descriptor and verified by proximity
    Nmu = []    #Number of features matched by descriptor but not verified (bad matches?)

    dv_mean = []    #mean distance of verified matches
    du_mean = []    #mean distance of unverified (bad?) matches
    dv_std = []     #standard deviation of distance of verified matches
    du_std = []     #standard deviation of distance of unverified matches
    
    for n in range(0,len(wins) ):
        
        #find features in next window
        win1 = wins[n]
        kps1 = detector.find_features(win1)
        kps1 = misc.filter_kps_border( kps1, win1.shape, border)
        print 'Image: ',n, 'Number of kps:' , len(kps1), 
        N.append(len(kps1))
        
        #match by proximity (to ensure same keypoints in both, used as max possible matches)
        matches = misc.proximity_match( kps0, kps1, dn)
        Nv.append( len(matches) )
        kps0_m = [ kps0[i] for i,j in matches]
        kps1_m = [ kps1[j] for i,j in matches] 
        print 'Proximity matched: ', Nv[-1], 
        
        #compute descriptors
        kps0_d, Ds0 = descriptor.compute_descriptors(win0, kps0_m)
        kps1_d, Ds1 = descriptor.compute_descriptors(win1, kps1_m)

        #match these two verified sets by descriptors
        ms = matcher.match(Ds1, Ds0)
        Nm.append(len(ms))
        print 'Descriptor matched: ', Nm[-1], 
        
        #verifiy each descriptor match is within tolerance dn specified.
        d_verified = []
        d_unverified = []
        for Dmatch in ms:
            
            if Dmatch.trainIdx==Dmatch.queryIdx:
                d_verified.append(Dmatch.distance )
            else:
                d_unverified.append(Dmatch.distance )
                
            #pull out matching keypoints
            #kp0 = kps0_d[Dmatch.trainIdx]
            #kp1 = kps1_d[Dmatch.queryIdx]

            #check proximity distance and store descriptor distance
            #pd = np.sqrt( (kp1.pt[0]-kp0.pt[0])**2 + (kp1.pt[1]-kp0.pt[1])**2 )
            #if pd<=dn:
            #    d_verified.append(Dmatch.distance )
            #else:
            #    d_unverified.append(Dmatch.distance )

        #store counts
        Nmv.append(  len(d_verified) )
        Nmu.append(  len(d_unverified) )
        
        #make arrays
        d_verified = np.array(d_verified)
        d_unverified = np.array(d_unverified)
        print 'verified: ', Nmv[-1], 'unverified: ', Nmu[-1],

        #calculate the stats and store results
        dv_mean.append(d_verified.mean())
        dv_std.append(d_verified.std())
        du_mean.append(d_unverified.mean())
        du_std.append(d_unverified.std())

        print 'dv mean: {0:0.2f} std: {1:0.2f}'.format( dv_mean[-1], dv_std[-1]),
        print 'du mean: {0:0.2f} std: {1:0.2f}'.format( du_mean[-1], du_std[-1])

    #convert to array types
    N = np.array(N, np.float64)       #number of features
    Nv = np.array(Nv, np.float64)     #number proximity matched
    Nm = np.array(Nm, np.float64)     #number of descriptor matched features
    Nmv = np.array(Nmv, np.float64)   #number of verified (i.e. also proximity matched) descriptor matched features
    Nmu = np.array(Nmu, np.float64)   #number of unverified (i.e. not also proximity matched) descriptor matched features

    #distances
    dv_mean = np.array(dv_mean)       #mean distance of verified matches
    du_mean = np.array(du_mean)       #mean distance of unverified matches
    dv_std = np.array(dv_std)         #std. dev. of distance of verified matches
    du_std = np.array(du_std)         #std. dev. of distance of unverified matches
    
    #store arrays to results
    res['N'] = N
    res['Nv'] = Nv
    res['Nm'] = Nm    
    res['Nmv'] = Nmv
    res['Nmu'] = Nmu
    res['dv_mean'] = dv_mean
    res['du_mean'] = du_mean
    res['dv_std'] = dv_std
    res['du_std'] = du_std
    
    #calculate robustness
    res['R'] = Nmv/Nv                  #robustness
    
    return res



#Paper method!
def robustness_match3( wins, detector, descriptor, matcher,  dn=1.5, ref_i=0, border=48):
    """
    Do robustness check on shifted windows (see get_shifted_windows).
    
    *** Calculate descriptors at keypoint found in first window  and try to match
        this ensures the same points are examined as windows are translation corrected
        and that orientation of keypoints is fixed                              ***
    
    1) Find keypoints in first image window
    2) Compute descriptors for these keypoints in both and match
    3) Divide matches into verified (Nmv) and unverified (Nmu) matches
        and find the mean and standard deviation of their descriptor distances.
    4) Compute R = Nmv / Nv

    Args:
    dn      -   promitity test tolerance
    ref_i   -   reference frame index
    border  -   exclude features inside the width border region
                (some descriptors excluded those to close to image edge)

    Returns results dictionary with keys: 
    method  -   method used
    N   -   Number of features found
    Nm  -   Number of features matched by descriptor
    Nmv -   Number of features matched by descriptor and verified by proximity
    Nmu -   Number of features matched by descriptor but not verified (bad matches?)

    dv_mean -   mean distance of verified matches
    du_mean -   mean distance of unverified matches
    dv_std  -   standard deviation of distance of verified matches
    du_std  -   standard deviation of distance of unverified matches
    
    R = Nmv/N - robustness
    """

    #find keypoints in original window v
    win0 = wins[ref_i]
    kps0 = detector.find_features(win0) 
    kps0 = misc.filter_kps_border( kps0, win0.shape, border)
    N0 = len(kps0)
    
    #compute descriptors for this reference set
    kps0_d, Ds0 = descriptor.compute_descriptors(win0, kps0)

    #Loop over all windows and match to first window found
    res = {}
    res['method'] = descriptor.method
    Nm = []     #Number of feature smatched by descriptor
    Nmv = []    #Number of verified feature matchs 
    Nmu = []    #Number of unverified feature matchs
    dv_mean = []    #mean distance of verified matches
    du_mean = []    #mean distance of unverified (bad?) matches
    dv_std = []     #standard deviation of distance of verified matches
    du_std = []     #standard deviation of distance of unverified matches
    
    for n in range(0,len(wins) ):

        #compute descriptors in new window
        win1 = wins[n]
        kps1_d, Ds1 = descriptor.compute_descriptors(win1, kps0)
        
        #match these two verified sets by descriptors
        ms = matcher.match(Ds1, Ds0)
        Nm.append(len(ms))
        print 'Descriptor matched: ', Nm[-1], 
        
        #verifiy each descriptor match (should have same indices)
        d_verified = []
        d_unverified = []
        for Dmatch in ms:
            if Dmatch.trainIdx==Dmatch.queryIdx:
                d_verified.append(Dmatch.distance )
            else:
                d_unverified.append(Dmatch.distance )

        #store counts
        Nmv.append(  len(d_verified) )
        Nmu.append(  len(d_unverified) )
        
        #make arrays
        d_verified = np.array(d_verified)
        d_unverified = np.array(d_unverified)
        print 'verified: ', Nmv[-1], 'unverified: ', Nmu[-1],

        #calculate the stats and store results
        dv_mean.append(d_verified.mean())
        dv_std.append(d_verified.std())
        du_mean.append(d_unverified.mean())
        du_std.append(d_unverified.std())

        print 'dv mean: {0:0.2f} std: {1:0.2f}'.format( dv_mean[-1], dv_std[-1]),
        print 'du mean: {0:0.2f} std: {1:0.2f}'.format( du_mean[-1], du_std[-1])

    #convert to array types
    Nm = np.array(Nm, np.float64)     #number of descriptor matched features
    Nmv = np.array(Nmv, np.float64)   #number of verified (i.e. also proximity matched) descriptor matched features
    Nmu = np.array(Nmu, np.float64)   #number of unverified (i.e. not also proximity matched) descriptor matched features

    #distances
    dv_mean = np.array(dv_mean)       #mean distance of verified matches
    du_mean = np.array(du_mean)       #mean distance of unverified matches
    dv_std = np.array(dv_std)         #std. dev. of distance of verified matches
    du_std = np.array(du_std)         #std. dev. of distance of unverified matches
    
    #store arrays to results
    res['N'] = N0
    res['Nm'] = Nm    
    res['Nmv'] = Nmv
    res['Nmu'] = Nmu
    res['dv_mean'] = dv_mean
    res['du_mean'] = du_mean
    res['dv_std'] = dv_std
    res['du_std'] = du_std
    
    #calculate robustness
    res['R'] = Nmv/N0                  #robustness
    
    return res


#%%-----------------------------------------------------------------------------
# setup samples names etc
#-------------------------------------------------------------------------------
tolerance = 1.5 #tolerance used in proximity matching
ref_i = 0       #reference to this image

rect0 = (0, 32, 256 ,960)   #window size to use
max_i = 50  #limit to valid shifts Ax

#surface preparation methods to process
surfaces = ( ('horizontal milling', (2.5,16,50)),
            ('flat lapping reaming', (0.55,1.6,10)),
            ('vertical milling', (2.5,16,50)),
            ('turning', (2.5,16,50)),
            ('grinding', (0.55,1.6,10))  )
            
#surfaces = ( ('horizontal milling', (16,)),)
  
#figure subset
#surfaces = ( ('turning', (50,)),
#            ('horizontal milling', (16,)),
#            ('flat lapping reaming', (1.6,)),
#            ('vertical milling', (16,)),)
#            
            
path = '..\\data\\dataset4 surface samples translation\\'

#store output results here
out_path ='..\\results\\5.5.2 descriptor_translation_robustness.hdf5'      

#%%-----------------------------------------------------------------------------
# set methods to test
#-------------------------------------------------------------------------------

#Need to use orientated detector for this!
orb_detector = detectors.ORB_Detector()      #oFAST
brisk_detector = detectors.BRISK_Detector()     #oFAST
surf_detector = detectors.SURF_Detector()     #oDoG
usurf_detector = detectors.USURF_Detector()     #DoG

usurf1 = descriptors.SURF( upright=True, extended=False)
surf1 = descriptors.SURF( upright=False, extended=False)
orb32 = descriptors.ORB(patchSize=32)
brisk32 = descriptors.BRISK(patchSize=32)

#matchers to use
L1_norm = cv.BFMatcher(cv.NORM_L1, crossCheck=True)
L2_norm = cv.BFMatcher(cv.NORM_L2, crossCheck=True)
Hamming = cv.BFMatcher(cv.NORM_HAMMING, crossCheck=True)

to_test = ( (usurf_detector, usurf1, L1_norm), 
            (surf_detector, surf1, L1_norm),   
            (orb_detector, orb32, Hamming),
            (brisk_detector, brisk32, Hamming), 
            )
#%%-----------------------------------------------------------------------------
# Loop over each sample
#-------------------------------------------------------------------------------
out = h5py.File(out_path, 'a')  
for surface, Rzs in surfaces:
    for Rz in Rzs:
        print '------------------------------ {0} {1} ------------------------------'.format(surface, Rz)

        #-------------------------------------------------------------------------------
        # load data from hdf5
        #-------------------------------------------------------------------------------
        filename = '{0} Rz={1} um.hdf5'.format(surface, Rz)
        print 'Loading data from :', path+filename
        data = h5py.File(path+filename, 'r')
        imgs = data['imgs']
        
        #-------------------------------------------------------------------------------
        # get shifted windows
        #-------------------------------------------------------------------------------
        #speckles move +Ax, use pixel border to allow slight Ay changes
        Ax,Ay,c, wins = misc.get_shifted_windows(imgs, rect0, ref_i, max_i)
        data.close()
        
        #-------------------------------------------------------------------------------
        # set up group to store processed results
        #-------------------------------------------------------------------------------
        surface_full = '{0} Rz={1} um'.format(surface, Rz)
        #store common items to the surface group
        common = {'Ax':Ax, 'Ay':Ay,'c':c, 'tolerance':tolerance }
        surface_group = misc.save_dict_to_hdf5( common , out, key=surface_full)

        #-------------------------------------------------------------------------------
        # run robustness test
        #-------------------------------------------------------------------------------
        for detector, descriptor, matcher in to_test:
            
            print 'Doing ', descriptor.method, detector.method
            
            #include unmatchable features (present only in 1 frame)
            #res = robustness_match( wins, detector, descriptor, matcher, tolerance, ref_i=0, border=32)

            #only use those present in both - as found by proximity test
            #res = robustness_match2( wins, detector, descriptor, matcher, tolerance, ref_i=0, border=32)
            
            #Use same keypoint locations and orientations thorughout
            res = robustness_match3( wins, detector, descriptor, matcher, tolerance, ref_i=0, border=32)
            
            #save to file
            misc.save_dict_to_hdf5(res, surface_group, res['method'])
out.close()

