"""
Python module implementing descriptor methods
"""
import numpy as np
import time
import cv2 as cv


#%%-----------------------------------------------------------------------------
# Python class interface wrapping feature descriptor methods to store/return data
#-------------------------------------------------------------------------------
class Descriptor():
    def __init__(self, name=''):
        """Base class"""
        self.method = name
        self.descriptor = None  #see implementations ...
        
        #internal storage of results
        self.dt = []   #time to cacluate desriptors
        self.N  = []   #number of kp descriptor calculated.
    
    def reset(self):
        """
        Clear all internally stored data
        """
        self.N = []         #number of kp descriptor calculated.
        self.dt = []        #time to cacluate desriptors
        
    def get_data(self):
        """
        Return results as dictionary
        """
        res = {}
        res['method'] = self.method
        res['N'] = np.array(self.N)
        res['dt'] = np.array(self.dt)
        return res   
        
    def compute_descriptors(self, img, kps):
        """
        Compute & time the descriptors
        """
        tic=time.clock()
        kps, des = self.descriptor.compute(img, kps)
        toc=time.clock()
        
        #store results
        self.dt.append(toc-tic)
        self.N.append(len(kps))
        
        return kps, des

#%%-----------------------------------------------------------------------------
# USURF_64/USURF_128/SURF_64/SURF_128
#-------------------------------------------------------------------------------
class SURF(Descriptor):
    def __init__(self, upright=False, extended=False):
        """
        Setup SURF descriptor/matching method.
        
        threshold - threshold used for feature detection 300-50,000
        distance_threshold - reject matches with distance greater than this
        norm - matcher distance method cv.NORM_L1, cv.NORM_L2
        """
        if upright:
            if extended:
                method = 'USURF (length 128 descriptor)'        
            else:
                method = 'USURF (length 64 descriptor)'       
        else:
            if extended:
                method = 'SURF (length 128 descriptor)'        
            else:
                method = 'SURF (length 64 descriptor)'       
        Descriptor.__init__(self, method)
        
        #detection args
        threshold = 30    #only need for detection
        nOctaves = 1      #1 no image rescaling 4 by default
        nOctaveLayers = 3 #3 (deafult opencv)
        # patchsize is 20x20 for surf methods
        
        #descriptor args
        self.descriptor = cv.xfeatures2d.SURF_create(threshold, nOctaves, nOctaveLayers ,extended, upright)

#%%-----------------------------------------------------------------------------
# ORB
#-------------------------------------------------------------------------------
class ORB(Descriptor):
    def __init__(self, patchSize=32):
        """
        Setup ORB Descriptor method.

        patchSize   -   descriptor patch size
        """
        method='ORB (patchSize = {0})'.format(patchSize)    
        Descriptor.__init__(self, method)

        cv.ocl.setUseOpenCL(False) #work around for orb bug - this needs to be called first
        
        #detector args
        fastThreshold  = 30 #set automatically
        nFeatures= 5000     #max number of features to find
        scaleFactor = 1.2   #pyramid scale factor
        firstLevel = 0      #zero from opencv docs
        nLevels = 1         #pyramid levels 1 as no scale changes
        edgeThreshold = 0   #size of border to not detect in
        scoreType = cv.ORB_FAST_SCORE   # ORB_HARRIS_SCORE or ORB_FAST_SCORE

        #description args
        WTA_K = 2           #he number of points that produce each element of the 
                            #oriented BRIEF descriptor. The default value 2 means the 
                            #BRIEF where we take a random point pair and compare their 
                            #brightnesses, so we get 0/1 response (index of brightest)

        self.descriptor = cv.ORB_create( nFeatures, scaleFactor, nLevels, edgeThreshold, firstLevel, 
                             WTA_K, scoreType, patchSize, fastThreshold)

#%%-----------------------------------------------------------------------------
# BRISK
#-------------------------------------------------------------------------------
class BRISK(Descriptor):
    def __init__(self, patchSize=32 ):
        """
        Setup BRISK matching method.
        
        patchSize   -   descriptor patch size
        
        """
        method='BRISK (patchSize = {0})'.format(patchSize)   

        Descriptor.__init__(self, method)
        
        #detection
        fastThreshold = 30
        octaves = 0      #detection octaves. Use 0 to do single scale.
        
        #description
        patternScale = patchSize/37.0 #scale to the pattern used for sampling the neighbourhood of a keypoint.
                                      #(1=37x37 from paper)
        self.descriptor = cv.BRISK_create(fastThreshold, octaves, patternScale)

#%%-----------------------------------------------------------------------------
# distance calculations
#-------------------------------------------------------------------------------
def hamming(des0,des1):
    """
    Calculate the Hamming distance between two lists of keypoints
    """
    assert len(des0) == len(des1)
    
    ds = []
    for n in xrange(len(des0)):
        D0 = des0[n]
        D1 = des1[n]
        d = np.unpackbits( np.bitwise_xor(D0, D1) ).sum()
        ds.append(d)
    ds = np.array(ds)
    return ds
    
def L1_norm(des0, des1):
    """
    Calcualte the L1 norm (sum of absolute differences)
    """
    assert len(des0) == len(des1)
    ds = abs(des1-des0).sum(axis=1)
    return ds

def L2_norm(des0, des1):
    """
    Calcualte the L2 norm (sum of square of the differences)
    """
    assert len(des0) == len(des1)
    ds = ((des1-des0)**2).sum(axis=1)
    return ds