"""
Python module implementing detector methods
"""
import numpy as np
import time
import cv2 as cv

#%%-----------------------------------------------------------------------------
# Python classes wrapping feature detector methods to store data and
# automatically adjust thresholds as necessary
#-------------------------------------------------------------------------------
class Detector():
    def __init__(self, name=''):
        """Base class"""
        self.method = name
        
        #store data internally
        self.kps = []       #list of last detected keypoints
        self.N = []         #list of number of features found
        self.dt = []        #list of time taken per call 
        
    def reset(self):
        """
        Clear all internally stored data
        """
        self.kps = []       #list of last detected keypoints
        self.N = []         #list of number of features found
        self.dt = []        #list of time taken per call 
        
    def get_data(self):
        """
        Return results as dictionary
        """
        res = {}
        res['N'] = np.array(self.N)
        res['dt'] = np.array(self.dt)
        return res
        
    def find_features(self, img):
        """
        Find features in the image
        
        ##Implemented in subclasses##
        """
        pass
        

class Harris_Detector(Detector):
    def __init__(self):
        """
        Create a Harris corner detector
        """
        Detector.__init__(self, name='Harris')

        qualityLevel =  0.01 #0-1 reject below this quality
        maxCorners = 5000
        minDistance = 1
        blockSize = 3
        useHarris = True
        Harris_k = 0.04
        self.cv_harris = cv.GFTTDetector_create(maxCorners, qualityLevel, 
                                    minDistance, blockSize, useHarris, Harris_k)

    def find_features(self, img):
        """
        Find features in the image
        """
        tic = time.clock()
        kps = self.cv_harris.detect(img)
        toc = time.clock()
        
        #store data internally
        self.dt.append( toc-tic)
        self.N.append( len(kps) )
        self.kps = kps
        return kps
    
class GFTT_Detector(Detector):
    def __init__(self):
        """
        Create a GFTT/Shi-Tomasi corner detector
        """
        Detector.__init__(self, name='GFTT/Shi-Tomasi')

        qualityLevel =  0.01 #0-1 reject below this quality
        maxCorners = 5000
        minDistance = 1
        blockSize = 3
        useHarris = False
        Harris_k = 0.04
        self.cv_gftt = cv.GFTTDetector_create(maxCorners, qualityLevel, 
                                    minDistance, blockSize, useHarris, Harris_k)

    def find_features(self, img):
        """
        Find features in the image
        """
        tic = time.clock()
        kps = self.cv_gftt.detect(img)
        toc = time.clock()
        
        #store data internally
        self.dt.append( toc-tic)
        self.N.append( len(kps) )
        self.kps = kps
        return kps
        
class FAST_Detector(Detector):
    def __init__(self):
        """
        Create a FAST detector
        """
        Detector.__init__(self, name='FAST')
        
        threshold = 255 #adjusted automatically to 70% image mean
        nonmaxSuppression = True
        self.cv_fast = cv.FastFeatureDetector_create( threshold, 
                            nonmaxSuppression, cv.FastFeatureDetector_TYPE_9_16)

    def find_features(self, img):
        """
        Find features in the image
        """
        #adjust fast threshold automatically to image
        threshold = int( img.mean() * 0.7)
        self.cv_fast.setThreshold(threshold)
        
        #time
        tic = time.clock()
        kps = self.cv_fast.detect(img)
        toc = time.clock()
        
        #store data internally
        self.dt.append( toc-tic)
        self.N.append( len(kps) )
        self.kps = kps
        return kps
        
class AGAST_Detector(Detector):
    def __init__(self):
        """
        Create a AGAST detector
        """
        Detector.__init__(self, name='AGAST')
        
        threshold = 255 #adjusted automatically to 70% image mean
        nonmaxSuppression = True
        self.cv_agast = cv.AgastFeatureDetector_create( threshold, nonmaxSuppression, 
                                            cv.AgastFeatureDetector_OAST_9_16)

    def find_features(self, img):
        """
        Find features in the image
        """
        #adjust fast threshold automatically to image
        threshold = int(img.mean() * 0.7)
        self.cv_agast.setThreshold(threshold)
        
        #time
        tic = time.clock()
        kps = self.cv_agast.detect(img)
        toc = time.clock()
        
        #store data internally
        self.dt.append( toc-tic)
        self.N.append( len(kps) )
        self.kps = kps
        return kps

class USURF_Detector(Detector):
    def __init__(self):
        """
        Create a USURF / Difference of Gaussians (DoG) detector
        """
        Detector.__init__(self, name='USURF')
        
        threshold =  35 #KK used variable threshold 300 - 50000
        nOctaves = 1        #KK =1 no image rescaling 4 by default
        nOctaveLayers = 3 #KK used three default is 1 the minimum?
        extended = False
        upright = True
        self.cv_surf = cv.xfeatures2d.SURF_create(threshold, nOctaves, nOctaveLayers ,extended, upright)

    def find_features(self, img):
        """
        Find features in the image
        """
        tic = time.clock()
        kps = self.cv_surf.detect(img)
        toc = time.clock()
        
        #store data internally
        self.dt.append( toc-tic)
        self.N.append( len(kps) )
        self.kps = kps
        return kps

#... orientated methods    
class SURF_Detector(Detector):
    def __init__(self):
        """
        Create a SURF / orientated Difference of Gaussians (DoG) detector.
        """
        Detector.__init__(self, name='SURF')
        
        threshold =  35 #KK used variable threshold 300 - 50000
        nOctaves = 1        #KK =1 no image rescaling 4 by default
        nOctaveLayers = 3 #KK used three default is 1 the minimum?
        extended = False
        upright = False
        self.cv_surf = cv.xfeatures2d.SURF_create(threshold, nOctaves, nOctaveLayers ,extended, upright)

    def find_features(self, img):
        """
        Find features in the image
        """
        tic = time.clock()
        kps = self.cv_surf.detect(img)
        toc = time.clock()
        
        #store data internally
        self.dt.append( toc-tic)
        self.N.append( len(kps) )
        self.kps = kps

        return kps

class ORB_Detector(Detector):
    def __init__(self):
        """
        Create a ORB /orientated FAST detector
        """
        Detector.__init__(self, name='ORB')
        
        nFeatures= 5000
        scaleFactor = 1.2   #pyramid scale factor
        nLevels = 1       #pyramid levels 1 as no scale changes
        edgeThreshold = 0 #size of border to not detect in
        firstLevel = 0  #zero from opencv docs
        #for descriptors
        WTA_K = 2       #he number of points that produce each element of the oriented BRIEF descriptor. The default value 2 means the BRIEF where we take a random point pair and compare their brightnesses, so we get 0/1 response (index of brightest)
        scoreType = cv.ORB_FAST_SCORE #ORB_FAST_SCORE used for ranking top points only
        patchSize = 31  #size of patch used to find BRIEF dscriptor
        fastThreshold = 50

        self.cv_orb = cv.ORB_create(nFeatures, scaleFactor, nLevels, edgeThreshold, firstLevel, WTA_K, scoreType, patchSize, fastThreshold)
        #work around for orb bug - this needs to be called first
        cv.ocl.setUseOpenCL(False)

    def find_features(self, img):
        """
        Find features in the image
        """
        #adjust fast threshold automatically to image
        threshold = int( img.mean() * 0.7)
        self.cv_orb.setFastThreshold(threshold)
        
        #time
        tic = time.clock()
        kps = self.cv_orb.detect(img)
        toc = time.clock()
        
        #store data internally
        self.dt.append( toc-tic)
        self.N.append( len(kps) )
        self.kps = kps

        return kps
        
class BRISK_Detector(Detector):
    def __init__(self):
        """
        Create a BRISK /orientated fast/agast like detector
        """
        Detector.__init__(self, name='BRISK')
        
        #detector created in find_features as threshold cannot be adjusted
        
    def find_features(self, img):
        """
        Find features in the image
        """
        #create detector
        octaves = 0         #detection octaves. Use 0 to do single scale.
        patternScale = 1.8  #apply this scale to the pattern used for sampling the neighbourhood of a keypoint.
                            #in the paper the pattern is approximately 1 = +-18px, 1.8=+-32
        
        #adjust fast threshold automatically to 70% image mean 
        threshold = int( img.mean() * 0.7) 
        brisk = cv.BRISK_create(threshold, octaves, patternScale)

        #time
        tic = time.clock()
        kps = brisk.detect(img)
        toc = time.clock()
        
        #store data internally
        self.dt.append( toc-tic)
        self.N.append( len(kps) )
        self.kps = kps
    
        return kps
