"""
Python module implementing combined methods
"""
import numpy as np
import time
import cv2 as cv

from .misc import get_transform

#%%-----------------------------------------------------------------------------
# Python classes wrapping feature tracking methods to store data and
# automatically adjust thresholds as necessary
#-------------------------------------------------------------------------------

#%%-----------------------------------------------------------------------------
# Matcher method classes to allow easier to read code and store data internally
#-------------------------------------------------------------------------------
class Method():
    def __init__(self, name=''):
        self.method = name
        
        #store results internally
        self.Axs = []
        self.Ays = []
        self.thetas = []

        #number of features
        self.N0s = []   #in img0
        self.N1s = []   #in img1
        self.Nms = []   #matched 
        self.Nfms = []  #matched and filtered
        
        self.ds = [] #mean distance of points used
        self.dts = [] #time to process

    def reset(self):
        self.Axs = []
        self.Ays = []
        self.thetas = []

        self.N0s = []
        self.N1s = []
        self.Nms = []
        self.Nfms = []
        
    def get_data(self):
        """
        Return dictionary of arrays
        """
        data = {}
        data['Axs'] = np.array(self.Axs)
        data['Ays'] = np.array(self.Ays)
        data['thetas'] = np.array(self.thetas)
        data['N0s'] = np.array(self.N0s)
        data['N1s'] = np.array(self.N1s)
        data['Nms'] = np.array(self.Nms)
        data['Nfms'] = np.array(self.Nfms)
        data['ds'] = np.array(self.ds)
        data['dts'] = np.array(self.dts)
        
        return data
        
class NCC(Method):
    def __init__(self):
        """
        Normalised circular correlation method.
        
        No rotation calc.
        q-values stored in self.ds
        """
        Method.__init__(self, 'circular normalised cross-correlation')
        
    def calc_shift(self, img0,img1):
        """
        Find the translation between frames using the circular normaliued cross-correlation
        """
        #start timing
        tic = time.clock()
        
        #correlate
        cc = self.normxcorr_circ(img0,img1)
        
        #find peak
        py,px,q = self.find_peak_gaussian_circ(cc)   
        Ax = (px-(cc.shape[0]//2))
        Ay = (py-(cc.shape[1]//2))
        
        #stop timing
        toc = time.clock()
        self.dts.append(toc-tic)
        
        self.Axs.append(Ax)
        self.Ays.append(Ay)
        
        
        #store q value in ds
        self.ds.append(q)
        
        return Ax,Ay
    
    def normxcorr_circ(self, f, t):
        """
        Normalised circular cross-correlation
        """
        #subtract mean from t
        tt = np.array(t, dtype=np.float)- t.mean() 
        
        #find sum of squares of t
        sum_t2 = (tt*tt).sum()
        
        #find sum of squares of f
        N = f.size
        ff = np.array(f, dtype=np.float)
        sum_f = ff.sum()         #sum of image values
        sum_f2 = (ff*ff).sum()    #sum of squares of image values

        #find normalisation factor
        normf = np.sqrt( (sum_f2 - (sum_f*sum_f)/N) )
        norm = 1.0/( ( normf * np.sqrt(sum_t2) )  )
        
        #calc cc
        F = np.fft.rfft2(f)
        T = np.fft.rfft2(t)
        CC = (F*T.conj())* norm
        cc = np.fft.irfft2(CC)
        cc = np.fft.fftshift(cc)
        
        return cc
    
    def find_peak_gaussian_circ(self, cc):
        """
        Find the peak using gaussian subpixel interpolation corrected for circular
        correlation
        cc      -   correlation
        """
        #find peak and qvalue
        p = cc.flatten().argmax()
        py = p//cc.shape[1]
        px = p%cc.shape[1]
        mean = abs(cc).mean()
        
        q = cc[py,px]/mean
        
        ##find the peak value and weighting
        ny,nx = cc.shape
        x0 = y0 = cc[py,px]
        w0 = ( (nx -abs(px - nx//2)) * (ny -abs(py -  ny//2)) )
        
        ##Do y fitting
        if py-1<0:#wrap around
            ym1 = cc[ny-1, px ]
        else: 
            ym1 = cc[py-1, px ]
        wm1 = ( (nx -abs(px - nx//2)) * (ny -abs(py-1 - ny//2)) )

        if py+1 >= cc.shape[0]: #wrap arround
            yp1 = cc[ 0, px ]
        else:
            yp1 = cc[py+1, px]
        wp1 = ( (nx -abs(px - nx//2)) * (ny -abs(py+1 - ny//2)) )

        #subtract mean before weighting
        y0  = (y0-mean)
        ym1 = (ym1-mean)*(float(w0)/wm1);
        yp1 = (yp1-mean)*(float(w0)/wp1);
        
        top = ( np.log(ym1) -np.log(yp1) )
        bot = (2*(np.log(ym1) + np.log(yp1) - 2*np.log(y0) ))
        if bot==0:
            sy = py
        else:
            sy = py +( top / bot) 

        ##Do x fitting
        if px-1<0:#wrap around
            xm1 = cc[ py, nx-1 ]
        else:
            xm1 = cc[ py, px-1 ]
        wm1 = ( (nx -abs(px-1 - nx//2)) * (ny -abs(py - ny//2)) )
        
        if px+1>=cc.shape[1]:#wrap around
            xp1 = cc[ py, 0 ]
        else:
            xp1 = cc[ py, px+1 ]
        wp1 = ( (nx -abs(px+1 - nx//2)) * (ny -abs(py - ny//2)) )
        
        #subtract mean before weighting 
        x0  = (x0-mean)
        xm1 = (xm1-mean)*(float(w0)/wm1);
        xp1 = (xp1-mean)*(float(w0)/wp1);

        top = ( np.log(xm1) -np.log(xp1) ) 
        bot = (2*(np.log(xm1) + np.log(xp1) - 2*np.log(x0) ))
        if bot==0:
            sx=px
        else:
            sx = px +( top/bot) 

        if np.isnan(sy):
            sy = py
            #print 'nan in peak fitting'
            
        if np.isnan(sx):
            sx = px
            #print 'nan in peak fitting'
            
        return sy,sx,q

class ORB(Method):
    def __init__(self, distance_threshold=64 ):
        """
        Setup ORB matching method.

        distance_threshold - reject matches with distance greater than this
        """
        Method.__init__(self, 'ORB')
        cv.ocl.setUseOpenCL(False) #work around for orb bug - this needs to be called first
        
        fastThreshold  = 30 #set automatically
        nFeatures= 5000     #max number of features to find
        scaleFactor = 1.2   #pyramid scale factor
        nLevels = 1         #pyramid levels 1 as no scale changes
        patchSize = 32      #descriptor patch size
        edgeThreshold = patchSize   #size of border to not detect in
        firstLevel = 0      #zero from opencv docs
        WTA_K = 2           #he number of points that produce each element of the 
                            #oriented BRIEF descriptor. The default value 2 means the 
                            #BRIEF where we take a random point pair and compare their 
                            #brightnesses, so we get 0/1 response (index of brightest)
        scoreType = cv.ORB_FAST_SCORE   # ORB_HARRIS_SCORE or ORB_FAST_SCORE

        self.orb = cv.ORB_create( nFeatures, scaleFactor, nLevels, edgeThreshold, firstLevel, 
                             WTA_K, scoreType, patchSize, fastThreshold)

        #create BFMatcher object
        self.matcher = cv.BFMatcher(cv.NORM_HAMMING, crossCheck=True)
        self.distance_threshold = distance_threshold # reject matches with distance greater than this

    def calc_shift_and_rotation(self, img0, img1, maxMatches=None):
        
        #start timing
        tic = time.clock()
        
        #find kps in img0
        self.orb.setFastThreshold( int(img0.mean() *0.7) ) #adjust fast threshold
        kp0, des0  = self.orb.detectAndCompute(img0, mask=None)
        self.kp0 = kp0
        #lmiit KPs
        
        #find kps in img1
        self.orb.setFastThreshold( int(img1.mean()*0.7) ) #adjust fast threshold
        kp1, des1 = self.orb.detectAndCompute(img1, mask=None)
        #print 'found {0}, {1} keypoints'.format(len(kp0),len(kp1))
     
        # Match descriptors.
        matches = self.matcher.match(des0,des1)
        
        # Sort them in the order of their distance.
        matches = sorted(matches, key = lambda x:x.distance) 
        
        #filter out those whoose distance is too great.
        fmatches = filter( lambda x:x.distance<self.distance_threshold, matches)
        #print 'matches {0}, filtered matches {1}'.format(len(matches), len(fmatches))

        #reduce number of matches used
        if maxMatches is not None:
            fmatches = fmatches[0:maxMatches]

        if len(fmatches)<=2:
            Ax,Ay,theta = np.NaN, np.NaN,np.NaN
            ds = []
        else:
            #make two lists of points
            ds = [match.distance for match in  fmatches ]
            pointsA = [kp0[match.queryIdx].pt for match in fmatches ]
            pointsB = [kp1[match.trainIdx].pt for match in fmatches ]
            Ax,Ay,theta = get_transform( pointsB, pointsA )
        
        #stop timing
        toc = time.clock()
        self.dts.append(toc-tic)
        
        #store results internally
        self.Axs.append(Ax)
        self.Ays.append(Ay)
        self.thetas.append(theta)
        
        self.N0s.append(len(kp0))
        self.N1s.append(len(kp1))
        self.Nms.append(len(matches))
        self.Nfms.append(len(fmatches))
        self.ds.append( np.mean(ds))
        
        return Ax,Ay,theta
        
    def set_nfeatures(self, n):
        """
        Set number of features to detect
        """
        self.orb.setMaxFeatures(n)
        
class BRISK(Method):
    def __init__(self, distance_threshold=64 ):
        """
        Setup BRISK matching method.
        
        distance_threshold - reject matches with distance greater than this
        """
        Method.__init__(self, 'BRISK')
        self.octaves = 0      #detection octaves. Use 0 to do single scale.
        self.patternScale = 1 #scale to the pattern used for sampling the neighbourhood of a keypoint.
                              #(1=37x37 from paper)
        #CV BRISK created for each frame to adjust fastThreshold!
        #create BFMatcher object
        self.matcher = cv.BFMatcher(cv.NORM_HAMMING, crossCheck=True)
        self.distance_threshold = distance_threshold # reject matches with distance greater than this

    def calc_shift_and_rotation(self, img0, img1, maxMatches=None):

        #Need to create a new brisk object to adjust threshold 
        # (slows things down a bit but not included in the timing calculations)
        fastThreshold = int(img0.mean() *0.7) #adjust fast threshold
        brisk1 = cv.BRISK_create(fastThreshold, self.octaves, self.patternScale)
        fastThreshold = int(img1.mean() *0.7) #adjust fast threshold
        brisk2 = cv.BRISK_create(fastThreshold, self.octaves, self.patternScale)   
        
        #start timing
        tic = time.clock()
        #find kps in img0
        kp0, des0 = brisk1.detectAndCompute(img0, mask=None)
        
        #find kps in img1
        kp1, des1 = brisk2.detectAndCompute(img1, mask=None)
        #print 'found {0}, {1} keypoints'.format(len(kp0),len(kp1))
    
        # Match descriptors.
        matches = self.matcher.match(des0,des1)
        
        # Sort them in the order of their distance.
        matches = sorted(matches, key = lambda x:x.distance) 
        #filter out those whoose distance is too great.
        fmatches = filter( lambda x:x.distance<self.distance_threshold, matches)
        print 'matches {0}, filtered matches {1}'.format(len(matches), len(fmatches))

        #reduce number of matches used
        if maxMatches is not None:
            fmatches = fmatches[0:maxMatches]
        
        if len(fmatches)<=2:
            Ax,Ay,theta = np.NaN, np.NaN,np.NaN
            ds=[]
        else:
            #make two lists of points
            ds = [match.distance for match in  fmatches ]
            pointsA = [kp0[match.queryIdx].pt for match in fmatches ]
            pointsB = [kp1[match.trainIdx].pt for match in fmatches ]
            Ax,Ay,theta = get_transform( pointsB, pointsA )
        
        #stop timing
        toc = time.clock()
        self.dts.append(toc-tic)
        
        #store results internally
        self.Axs.append(Ax)
        self.Ays.append(Ay)
        self.thetas.append(theta)
        
        self.N0s.append(len(kp0))
        self.N1s.append(len(kp1))
        self.Nms.append(len(matches))
        self.Nfms.append(len(fmatches))
        
        self.ds.append( np.mean(ds))

        return Ax,Ay,theta
        
class USURF(Method):
    def __init__(self, distance_threshold=64, norm=cv.NORM_L1 ):
        """
        Setup USURF matching method.
        
        threshold - threshold used for feature detection 300-50,000
        distance_threshold - reject matches with distance greater than this
        norm - matcher distance method cv.NORM_L1, cv.NORM_L2
        """
        Method.__init__(self, 'USURF')
        
        threshold = 30      #seems to work ok
        nOctaves = 1      #KK =1 no image rescaling 4 by default
        nOctaveLayers = 4 #2 default
        extended = False    
        upright = True
        self.usurf = cv.xfeatures2d.SURF_create(threshold, nOctaves, nOctaveLayers ,extended, upright)

        #create BFMatcher object
        self.matcher = cv.BFMatcher(norm, crossCheck=True)
        self.distance_threshold = distance_threshold # reject matches with distance greater than this

    def calc_shift_and_rotation(self, img0, img1, maxMatches=None):
        #start timing
        tic = time.clock()
        
        #find kps in img0 and img1
        kp0, des0 = self.usurf.detectAndCompute(img0, mask=None)
        
        #find kps in img1
        kp1, des1 = self.usurf.detectAndCompute(img1, mask=None)
        #print 'found {0}, {1} keypoints'.format(len(kp0),len(kp1))
    
        # Match descriptors.
        matches = self.matcher.match(des0,des1)
        
        # Sort them in the order of their distance.
        matches = sorted(matches, key = lambda x:x.distance) 
        #filter out those whoose distance is too great.
        fmatches = filter( lambda x:x.distance<self.distance_threshold, matches)
        #print 'matches {0}, filtered matches {1}'.format(len(matches), len(fmatches))

        #reduce number of matches used
        if maxMatches is not None:
            fmatches = fmatches[0:maxMatches]
        
        if len(fmatches)<=2:
            Ax,Ay,theta = np.NaN, np.NaN,np.NaN
            ds=[]
        else:
            #make two lists of points
            ds = [match.distance for match in  fmatches ]
            pointsA = [kp0[match.queryIdx].pt for match in fmatches ]
            pointsB = [kp1[match.trainIdx].pt for match in fmatches ]
            Ax,Ay,theta = get_transform( pointsB, pointsA )
        
        #stop timing
        toc = time.clock()
        self.dts.append(toc-tic)
        
        #store results internally
        self.Axs.append(Ax)
        self.Ays.append(Ay)
        self.thetas.append(theta)
        
        self.N0s.append(len(kp0))
        self.N1s.append(len(kp1))
        self.Nms.append(len(matches))
        self.Nfms.append(len(fmatches))
        self.ds.append( np.mean(ds))

        return Ax,Ay,theta

class SURF(Method):
    def __init__(self, distance_threshold=64, norm=cv.NORM_L1 ):
        """
        Setup SURF matching method.
        
        distance_threshold - reject matches with distance greater than this
        norm - matcher distance method cv.NORM_L1, cv.NORM_L2
        """
        Method.__init__(self, 'USURF')
        
        threshold=30       #300-50,000
        nOctaves = 1        #KK =1 no image rescaling 4 by default
        nOctaveLayers = 4 #KK used three default is 1 the minimum?
        extended = False
        upright = False
        self.surf = cv.xfeatures2d.SURF_create(threshold, nOctaves, nOctaveLayers ,extended, upright)

        #create BFMatcher object
        self.matcher = cv.BFMatcher(norm, crossCheck=True)
        self.distance_threshold = distance_threshold # reject matches with distance greater than this

    def calc_shift_and_rotation(self, img0, img1, maxMatches=None):
        #start timing
        tic = time.clock()
        
        #find kps in img0 and img1
        kp0, des0 = self.surf.detectAndCompute(img0, mask=None)
        
        #find kps in img1
        kp1, des1 = self.surf.detectAndCompute(img1, mask=None)
        #print 'found {0}, {1} keypoints'.format(len(kp0),len(kp1))
    
        # Match descriptors.
        matches = self.matcher.match(des0,des1)
        
        # Sort them in the order of their distance.
        matches = sorted(matches, key = lambda x:x.distance) 
        #filter out those whoose distance is too great.
        fmatches = filter( lambda x:x.distance<self.distance_threshold, matches)
        #print 'matches {0}, filtered matches {1}'.format(len(matches), len(fmatches))

        #reduce number of matches used
        if maxMatches is not None:
            fmatches = fmatches[0:maxMatches]
        
        if len(fmatches)<=2:
            Ax,Ay,theta = np.NaN, np.NaN,np.NaN
            ds=[]
        else:
            #make two lists of points
            ds = [match.distance for match in  fmatches ]
            pointsA = [kp0[match.queryIdx].pt for match in fmatches ]
            pointsB = [kp1[match.trainIdx].pt for match in fmatches ]
            Ax,Ay,theta = get_transform( pointsB, pointsA )
        
        #stop timing
        toc = time.clock()
        self.dts.append(toc-tic)
        
        #store results internally
        self.Axs.append(Ax)
        self.Ays.append(Ay)
        self.thetas.append(theta)
        
        self.N0s.append(len(kp0))
        self.N1s.append(len(kp1))
        self.Nms.append(len(matches))
        self.Nfms.append(len(fmatches))
        self.ds.append( np.mean(ds))
        
        return Ax,Ay,theta
    
