"""
Python module containing miscelaneous functions/classes used in multiple places
"""
import numpy as np
import cv2 as cv

#%%-----------------------------------------------------------------------------
# peak finding functions
#-------------------------------------------------------------------------------
def find_peak_gaussian(cc):
    """
    Find the peak using 3-point gaussian subpixel interpolation
    cc      -   correlation
    """
    p = cc.flatten().argmax()
    py = p//cc.shape[1]
    px = p%cc.shape[1]

    q = cc[py,px]/abs(cc.mean())

    y0 = x0 = cc[py,px]
    if py-1<0:
        ym1 = cc[ py+1, px ]
    else:
        ym1 = cc[py-1, px ]

    if py+1 >= cc.shape[0]:
        yp1 = cc[ py-1, px ]
    else:
        yp1 = cc[py+1, px]

    if px-1<0:
        xm1 = cc[ py, px+1 ]
    else:
        xm1 = cc[ py, px-1 ]

    if px+1>=cc.shape[1]:
        xp1 = cc[ py, px-1 ]
    else:
        xp1 = cc[ py, px+1 ]

    top = ( np.log(ym1) -np.log(yp1) )
    bot = (2*(np.log(ym1) + np.log(yp1) - 2*np.log(y0) ))
    if bot==0:
        sy = py
    else:
        sy = py +( top / bot) 

    top = ( np.log(xm1) -np.log(xp1) ) 
    bot = (2*(np.log(xm1) + np.log(xp1) - 2*np.log(x0) ))
    if bot==0:
        sx=px
    else:
        sx = px +( top/bot) 

    if np.isnan(sy):
        sy = py
        #print 'nan in peak fitting'
        
    if np.isnan(sx):
        sx = px
        #print 'nan in peak fitting'

    return sy,sx,q
  
#%%-----------------------------------------------------------------------------
# key point helper functions
#-------------------------------------------------------------------------------
def proximity_match( kps0, kps1, dn=1):
    """
    Match kps with the tolerance dn given of each other
    
    Returns list of the matching indices (i,j) in kps0, kps1 so that 
    kps0[i] = kps1[j]
    """
    #copy
    kps1 = list(kps1)

    matches = []
    
    #check each kps0 for match to within +- dn pixels in this rotated set
    for i,kp0 in enumerate(kps0):
        match = None
    
        #calculate (index, distances) to all kp in kps1
        distance = np.sqrt(((cv.KeyPoint_convert(kps1) - (kp0.pt[0], kp0.pt[1]))**2).sum(axis=1))
        
        #find closest
        j = distance.argmin()
        
        #check distance
        if distance[j]<=dn:
            matches.append( (i,j) )
        
    return matches
    
def filter_kps( kps, centre=(200,200), radius=100):
    """
    Filter key points to limit to within a circular window.
    """
    Cy,Cx  = centre
    kps_ = [ kp for kp in kps if (np.hypot(kp.pt[0]-Cx, kp.pt[1]-Cy)<radius)]
    return kps_
    
def filter_kps_border( kps, shape, border):
    """
    Filter key points to remove any in image border
    """
    #print len(kps),
    kps_ = [  kp  for kp in kps if ( (kp.pt[0]>border) and (kp.pt[0]<shape[1]-border) and (kp.pt[1]>border) and (kp.pt[1]<shape[0]-border) ) ]
    #print len(kps_)
    return kps_

def rotate_kps( kps, Dang, centre=(200,200)):
    """
    Rotate a set of keypoints
    
    *** +ve angle is clockwise due to image axis coordinate system ***
    """
    Cy,Cx  = centre
    # x_ = x*cos(Dang) - y*sin(Dang)
    # y_ = x*sin(Dang) + y*cos(Dang)
    sinDang = np.sin(np.radians(Dang))
    cosDang = np.cos(np.radians(Dang))
    
    kps_ = []
    for kp in kps:
        x,y = kp.pt
        x_ = ((x-Cx)*cosDang - (y-Cy)*sinDang) + Cx
        y_ = ((x-Cx)*sinDang + (y-Cy)*cosDang) + Cy
        #create a copy of the kp with new x,y
        ang_ = (kp.angle+Dang)%360
        kp_ = cv.KeyPoint( x_, y_, kp.size, ang_, kp.response, kp.octave, kp.class_id)
        kps_.append(kp_)
    return kps_

#%%-----------------------------------------------------------------------------
# Functions to generate aligned windows in data sets for analysis
#-------------------------------------------------------------------------------
def get_shifted_windows( imgs, rect0=(0,0,128,128), ref_i=0, max_i = None):
    """
    Get a set of windows with the translation removed. 
    
    Starting from the window defined by rect0=(x0,y0,w,h) in imgs[ref_i] find and 
    return the same window in the following images corrected for the 
    translation using normalised cross-correlation
    
    Returns:
    Ax(t), Ay(t)        -   the translation found from reference
    c(t)                -   the normalise cross-correlation coefficent
    wins(t,y,x)         -   the shifted windows
    """
    #window in first image
    x0,y0,w,h = rect0
    y1 = y0+h
    x1 = x0+w
    img0 = imgs[ref_i]
    win0 = img0[y0:y1, x0:x1]
    h,w = win0.shape
    
    cs = []         #correlation coefficent (measure of speckle deformation)
    Axs = []        #shift between speckles
    Ays = []        #"  "
    wins = []
    
    if max_i is None: 
        max_i = imgs.shape[0]
    
    for n in range(0,max_i):
        
        #find shift using normalised cross-correlation
        cc = cv.matchTemplate(imgs[n],win0, cv.TM_CCOEFF_NORMED)
        
        py,px,q = find_peak_gaussian(cc)
        Ay = py - y0
        Ax = px - x0
        print 'Ax, Ay, q', Ax, Ay ,q,
        
        #store to lists
        Axs.append(Ax)
        Ays.append(Ay)
        
        #Adjust the window position to ensure all keypoints should still be present
        win1 = imgs[n, y0+int(np.rint(Ay)):y1+int(np.rint(Ay)), x0+int(np.rint(Ax)):x1+int(np.rint(Ax))]
        #print 'win1', y0+int(np.rint(Ay)),y1+int(np.rint(Ay)), x0+int(np.rint(Ax)),x1+int(np.rint(Ax)), win1.shape

        wins.append(win1)

        #find the decorrelation/correlation coefficent value at center pixel
        cc = cv.matchTemplate(win0,win1, cv.TM_CCOEFF_NORMED)
        cor = cc[cc.shape[0]//2,cc.shape[1]//2]
        cs.append( cor )
        print 'correlation coefficent: ',cor
    
    #convert to array type for ease
    cs = np.array(cs, np.float64)
    Axs = np.array(Axs, np.float64)
    Ays = np.array(Ays, np.float64)
    wins = np.array(wins)
    
    return Axs,Ays,cs, wins
 
def get_rotated_windows( imgs, angles, centre=(512,640), radius=100, pad=32, ref_i=0, mask=False):
    """
    Returns windows centred on the rotation with the centre and radius given.
    
    Starting from the window defined by centre and radius in imgs[0] find and 
    return the same window in the following images corrected for the known 
    rotation and any remaining translation using normalised cross-correlation
    
    Also estimate the decorrelation of the window by rotating by the known angle
    and computing the normalised cross-correlation co-efficent.
    
    Method:
    1) get window in reference image (win0)
    2) rotate new image (img1) to reference frame (img1')
    3) Find win0 translation in img1'
    4) Adjust window location in img1' and store (rotation and translation corrected)
    5) Adjust window loctaion in img1 and store (only translation corrected)
    
    imgs    -   images array
    angles  -   angles
    centre  -   approximate centre of rotation
    radius  -   radius of window to use
    pad     -   extra border padding
    ref_i   -   reference index
    mask    -   Apply a circular mask
    
    Returns:
    
    Axs,Ays     -   translation found
    Dangs       -   angle change compared to reference
    cs          -   normalised cross-correlation co-efficent of windows
    wins        -   Windows centered on rotation (translation corrected)
    wins_       -   Windows rotated and centered to match reference window.
    """
    Cy,Cx = centre

    #indices for window
    x0, y0 = Cx-radius-pad , Cy-radius-pad
    x1, y1 = Cx+radius+pad , Cy+radius+pad
    
    #reference window
    img0 = imgs[ref_i]
    ang0 = angles[ref_i]
    h,w = img0.shape
    win0 = img0[y0: y1,  x0: x1]

    #create circular mask to exclude speckles not found in all!
    y,x = np.indices(win0.shape)
    x = x - win0.shape[1]//2
    y = y - win0.shape[0]//2
    r = np.sqrt(x**2 + y**2)
    cmask = np.ones( r.shape, dtype=np.uint8)
    cmask[r >radius ] = 0
    
    cs = []         #correlation coefficent (measure of speckle deformation)
    wins = []       #centred but still rotated
    wins_ = []      #rotation removed
    Axs = []        #shift of window 
    Ays = []        # ""
    Dangs = []       #rotation from reference.
    
    for n in range(0,imgs.shape[0]):
        
        #rotate image by known angle compared to reference
        ang1 = angles[n]
        Dang = ang1-ang0
        Dangs.append(Dang)
        print 'Rotation: ',Dang,
    
        M = cv.getRotationMatrix2D( (Cx,Cy), -Dang ,1)
        img1_ = cv.warpAffine(imgs[n], M, (w,h) ) #shape arg is w,h not h,w!
    
        #find any translation using normalised cross-correlation
        cc = cv.matchTemplate(img1_,win0, cv.TM_CCOEFF_NORMED)
        
        py,px,q = find_peak_gaussian(cc)
        Ay_ = py - y0
        Ax_ = px - x0
        print 'Ax_, Ay_, q', Ax_, Ay_ ,q, 
        
        #Adjust the window position to ensure all keypoints should still be present
        # -- windows have same rotation and position. --
        win1_ = img1_[y0+int(np.rint(Ay_)):y1+int(np.rint(Ay_)), x0+int(np.rint(Ax_)):x1+int(np.rint(Ax_))]

        #find the decorrelation/correlation coefficent value at center pixel
        cc = cv.matchTemplate(win0,win1_, cv.TM_CCOEFF_NORMED)
        cor = cc[cc.shape[0]//2,cc.shape[1]//2]
        print 'correlation coefficent: ',cor,
        
        # -- windows have same centre position different rotation --
        #calculate Ax,Ay in original image1
        Dang_rad = np.radians(-Dang)
        Ax = Ax_*np.cos(Dang_rad) - Ay_*np.sin(Dang_rad)
        Ay = Ay_*np.cos(Dang_rad) + Ax_*np.sin(Dang_rad)
        print 'Ax, Ay', Ax, Ay
        win1 = imgs[n, y0+int(np.rint(Ay)):y1+int(np.rint(Ay)), x0+int(np.rint(Ax)):x1+int(np.rint(Ax))]

        #apply circular mask
        if mask:
            win1 = win1*cmask
            win1_ = win1_*cmask
        
        #store to lists
        Axs.append(Ax_)
        Ays.append(Ay_)
        cs.append(cor)
        wins.append(win1)
        wins_.append(win1_)

    #convert to array type for ease
    cs = np.array(cs, np.float64)
    Axs = np.array(Axs, np.float64)
    Ays = np.array(Ays, np.float64)
    Dangs = np.array(Dangs, np.float64)
    wins = np.array(wins)
    wins_ = np.array(wins_)

    return Axs,Ays, Dangs, cs, wins, wins_

#%%-----------------------------------------------------------------------------
# load/save to hdf5 helpers
#-------------------------------------------------------------------------------
class Structure():
    def __init__(self, dictlike):
        """Structure like interface to dictionary like objects"""
        self.__dict__ = dictlike

#save dict
def save_dict_to_hdf5( dict, hdf5, key='' ):
    """
    Save the dictionary to the hdf5 group/file
    """
    if key is not '':
        if key in hdf5:
            group = hdf5[key]
        else:
            group = hdf5.create_group(key)
    else:
        group = hdf5

    for k,v in dict.items():
        if type(v) in [str, unicode, int, float, complex]:
            #store to attribute rather than data set
            if k in group.attrs:
                group.attrs.pop(k)
            #store new attribute
            group.attrs[k] = v
        else:
            #store as data set
            if k in group:  #remove if already exisits
                group.pop(k)
            #store new vlaue
            group[k]=v
    return group
    
def load_dict_from_hdf5( hdf5, key=''):
    """
    Load a dictionary from the hfd5 group/file
    """
    if key is not '':
        group = hdf5[key] 
    else:
        group = hdf5

    res = {}
    for k,v in group.items():
        res[k]=v[:]
        
    for k, v in group.attrs.items():
        res[k] = v
        
    return res

#%%-----------------------------------------------------------------------------
# Rigid transform calculations with/without centre of rotation calculation
#-------------------------------------------------------------------------------
def get_transform( pointsA, pointsB ):
    """
    Calculate the transformation between two frames from a series of points
    identified in each.
    
    returns dx,dy, theta
    """
    if len(pointsB)!=len(pointsA):
        raise Exception('Expected equal number of points A and B')
    
    # 1) Find the sums needed for least squares solution
    # S1 = sum( xi**2 + yi**2 )
    # S2 = sum( xi )
    # S3 = sum ( yi )
    # NC  = number of points
    # Sk1 = sum( wi*(xi*kxi + yi*kyi) )
    # Sk2 = sum( wi*(xi*kyi - yi*kxi) )
    # Sk3 = sum( wi*(kxi) )
    # Sk4 = sum( wi*(kyi) )
    # Use longs to prevent overflows
    S1 = S2 = S3 = long(0)
    Sk1 = Sk2 = Sk3 = Sk4 = long(0)
    NC = len(pointsA)
    for i in xrange(0,NC):
        x,y = pointsA[i]
        
        S1 += (x**2 + y**2)
        S2 += x
        S3 += y
        
        kx,ky = pointsB[i]
        Sk1 += (x*kx + y*ky)
        Sk2 += (x*ky - y*kx)
        Sk3 += kx
        Sk4 += ky

    #print 'S1=',S1
    #print 'S2=',S2
    #print 'S3=',S3
    #print 'Sk1=',Sk1
    #print 'Sk2=',Sk2
    #print 'Sk3=',Sk3
    #print 'Sk4=',Sk4
    #print 'NC=',NC
    
    # 2) calculate some other values used in muliple times
    K1 = (NC*S1 - S2**2 - S3**2)
    #print 'K1=',K1
    
    denom = 1.0/K1
    #print 'denom=',denom
    
    K2 = (S2**2 + S3**2 + K1)/S1
    #print 'K2=',K2
    
    #X = denom * Matrix([
    #X = denom * sp.Matrix([
    #            [( -S2*Sk3 - S3*Sk4 + Sk1*K2 ) ],
    #            [( -S2*Sk4 + S3*Sk3 + Sk2*K2 ) ],
    #            [     S1*Sk3 - S2*Sk1 + S3*Sk2 ],
    #            [     S1*Sk4 - S2*Sk2 - S3*Sk1 ]])
    #where: 
    # X = sp.Matrix( [cos, sin, dx, dy])

    cos_theta = denom*( -S2*Sk3 - S3*Sk4 + Sk1*K2 )
    sin_theta = denom*( -S2*Sk4 + S3*Sk3 + Sk2*K2 )
    dx = denom*( S1*Sk3 - S2*Sk1 + S3*Sk2 )
    dy = denom*(S1*Sk4 - S2*Sk2 - S3*Sk1 )

    theta = np.degrees(np.arctan2(sin_theta,cos_theta))
    
    
    return dx, dy, theta

def get_transform2( pointsA, pointsB ):
    """
    Calculate the transformation between two frames from a series of points
    identified in each. Also calculates the centre of rotation.
    
    returns dx,dy, theta , Cx,Cy
    """
    if len(pointsB)!=len(pointsA):
        raise Exception('Expected equal number of points A and B')
    
    # 1) Find the sums needed for least squares solution
    # S1 = sum( xi**2 + yi**2 )
    # S2 = sum( xi )
    # S3 = sum ( yi )
    # NC  = number of points
    # Sk1 = sum( wi*(xi*kxi + yi*kyi) )
    # Sk2 = sum( wi*(xi*kyi - yi*kxi) )
    # Sk3 = sum( wi*(kxi) )
    # Sk4 = sum( wi*(kyi) )
    S1 = S2 = S3 = long(0)
    Sk1 = Sk2 = Sk3 = Sk4 = long(0)
    NC = len(pointsA)
    for i in xrange(0,NC):
        x,y = pointsA[i]
        
        S1 += (x**2 + y**2)
        S2 += x
        S3 += y
        
        kx,ky = pointsB[i]
        Sk1 += (x*kx + y*ky)
        Sk2 += (x*ky - y*kx)
        Sk3 += kx
        Sk4 += ky

    #print 'S1=',S1
    #print 'S2=',S2
    #print 'S3=',S3
    #print 'Sk1=',Sk1
    #print 'Sk2=',Sk2
    #print 'Sk3=',Sk3
    #print 'Sk4=',Sk4
    #print 'NC=',NC
    
    # 2) calculate some other values used in muliple times
    K1 = (NC*S1 - S2**2 - S3**2)
    #print 'K1=',K1
    
    denom = 1.0/K1
    #print 'denom=',denom
    
    K2 = (S2**2 + S3**2 + K1)/S1
    #print 'K2=',K2
    
    #X = denom * Matrix([
    #X = denom * sp.Matrix([
    #            [( -S2*Sk3 - S3*Sk4 + Sk1*K2 ) ],
    #            [( -S2*Sk4 + S3*Sk3 + Sk2*K2 ) ],
    #            [     S1*Sk3 - S2*Sk1 + S3*Sk2 ],
    #            [     S1*Sk4 - S2*Sk2 - S3*Sk1 ]])
    #where: 
    # X = sp.Matrix( [cos, sin, dx, dy])
    cos_theta = denom*( -S2*Sk3 - S3*Sk4 + Sk1*K2 )
    sin_theta = denom*( -S2*Sk4 + S3*Sk3 + Sk2*K2 )
    
    dx = denom*(S1*Sk3 - S2*Sk1 + S3*Sk2 ) #need negative here to correct for
    dy = denom*(S1*Sk4 - S2*Sk2 - S3*Sk1 ) #... image coords.

    theta_r = np.arctan2(sin_theta,cos_theta)
    theta =  np.degrees(theta_r) 
    
    # 3) find centre of rotation
    #    C = sp.Matrix( ('Cx','Cy'))
    #    T = sp.Matrix( ('dx','dy'))
    #    I = sp.Matrix( ((1,0),(0,1)) )
    #    theta = sp.symbols('theta')
    #    R = sp.Matrix( ( (sp.cos(theta) , -sp.sin(theta)),(sp.sin(theta), sp.cos(theta)) ) )
    #    From at C=(Cx,Cy) then total translation is just from T
    #    C = RC + T
    #    C-RC = T
    #    C(I-R) = T
    #    C = (I-R).inv() * T
    if cos_theta == 1:
        Cx = None
        Cy = None
    else:
        #recalculate cos_theta/sin_theta for better stability
        cos_theta = np.cos(-theta_r)    
        sin_theta = np.sin(-theta_r)
        
        #find centre of rotation
        Cx = dx*( 1/(-cos_theta + 1) - sin_theta**2 /(2*(-cos_theta + 1)**2) ) \
             -dy*sin_theta/(2*(-cos_theta + 1))
             
        Cy = (dx*sin_theta)/(2*(-cos_theta + 1)) + dy/2

        #translate (0,0) to (Cx,Cy) frame
        x0 = -Cx        
        y0 = -Cy
        #print 'x0,y0 : ',(x0,y0)
        
        #find shifted (0,0) point in (Cx,Cy) frame
        x1 = cos_theta*x0 - sin_theta*y0    
        y1 = sin_theta*x0 + cos_theta*y0
        #print 'x1,y1 : ',(x1,y1)
        
        #find shift at (0,0) due to theta
        dx_rot = x1-x0  
        dy_rot = y1-y0

        #correct dx,dy
        #print 'dx,dy',(dx,dy)
        #print 'dx,dy rot: ',(dx_rot, dy_rot)
        dx = dx - dx_rot
        dy = dy - dy_rot
        #print (dx,dy)
        
    return dx, dy, theta, Cx,Cy

def generate_test_points( shape=(512,512), skip=10, T=(0,0), R=5.0, C=(256,256)):
    """
    Generate list of points before/after transformation for testing
    """
    x,y = np.meshgrid( range(0,shape[1],skip), range(0, shape[0],skip))
    y = y.flatten()
    x = x.flatten()
 
    sin = np.sin(np.radians(R))
    cos = np.cos(np.radians(R))
    
    Tx,Ty = T
    Cx,Cy = C
    
    x_ = ((x-Cx)*cos - (y-Cy)*sin) + Cx + Tx
    y_ = ((x-Cx)*sin + (y-Cy)*cos) + Cy + Ty
    
    return zip(x,y), zip(x_,y_)
    
#%%-----------------------------------------------------------------------------
# Modelling translations/rotations
#-------------------------------------------------------------------------------
def model_shifted_windows(ref_img, win0=(0,0,128,128), delta=(0,0), M=10):
    """
    Return a shifted window in the image for subpixel shifts in 1/M px steps
    
    win0 =(x0,y0,w,h) must allow for 1px border in calculations
    delta = (dx, dy) in pixels
    """
    m=1.0/M
    
    #initial window location
    x0,y0,w0,h0 = win0    
    img0 = ref_img[y0:y0+h0, x0:x0+w0]
    
    #split shift into whole pixels and subpixels
    delta = np.rint(np.array(delta)*M)
    #print delta
    dx = int((delta[0])//M )    #whole pixels
    dy = int((delta[1])//M )
    sdx = int((delta[0])%M )          #sub-pixels
    sdy = int((delta[1])%M )
    #print 'dx,dy,sdx,sdy', dx,dy,sdx,sdy
    
    #find shifted window location to whole pixels   
    #scale ref_img up using nearest neighbour - only for a region around the new
    #shifted window (a 1px border around window to account for subpixel part of shift)
    x0 = x0+dx -1
    y0 = y0+dy -1
    x1 = x0+w0+2
    y1 = y0+h0+2
    #print 'x0,y0,x1,y1', x0,y0,x1,y1
    #   cv.INTER_NEAREST 
    #   cv.INTER_LINEAR 
    #   cv.INTER_AREA
    ref_img_M = cv.resize(ref_img[y0:y1, x0:x1], dsize=None, fx=M, fy=M, interpolation=cv.INTER_NEAREST)
    
    #find the subpixel shifted window inside the magnified image
    Mx0 = M + sdx   #window starts M pixels in now, then shift by the subpixels amount
    My0 = M + sdy
    Mx1 = Mx0 + (w0*M)
    My1 = My0 + (h0*M)
    #print 'Mx0,My0, Mx1,My1', Mx0,My0,Mx1,My1
    
    #find the new window location
    img1_M = ref_img_M[ My0:My1, Mx0:Mx1]
    
    #scale the shifted image down
    img1 = cv.resize(img1_M, dsize=None, fx=m, fy=m, interpolation=cv.INTER_AREA)

    return np.ascontiguousarray(img0), np.ascontiguousarray(img1)

def model_rotated_window(ref_img, shape=(256,256), ang=0.0):
    """
    Return a rotated window in the reference speckle pattern.
    
    returns img0 (un-rotated) and img1 (rotated)
    """
    w,h = shape
    rows, cols = ref_img.shape
    M = cv.getRotationMatrix2D((cols/2,rows/2),ang,1)
    imgR = cv.warpAffine(ref_img,M,(cols,rows), flags=cv.INTER_CUBIC, borderMode=cv.BORDER_CONSTANT, borderValue=0)
    
    x0 = (cols/2) - (w/2)
    y0 = (rows/2) - (h/2)

    img0 = ref_img[y0:y0+h,x0:x0+w]
    img1 = imgR[y0:y0+h,x0:x0+w]

    return  np.ascontiguousarray(img0), np.ascontiguousarray(img1)
